# Third party modules
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import matplotlib.pyplot as plt
import numpy as np
import tkinter as tk
from tkinter import messagebox

class MaterialParamSelector:
    def __init__(
            self, root, x, y, tth, preselected_materials, label, on_complete):

        self.root = root
        self.root.title('Material Parameter Selection')
        self.on_complete = on_complete  # Completion callback

        # Reference data
        self.ref_data_x = x
        self.ref_data_y = y
        self.ref_data_label = label
        self.tth = tth

        # Materials
        self.materials = []
        self.selected_material = None

        # Create plot
        self.figure, self.ax = plt.subplots()
        self.legend_handles = []
        self.canvas = FigureCanvasTkAgg(self.figure, self.root)
        self.canvas.get_tk_widget().grid(row=0, column=0, rowspan=12)

        # Widgets for material list and parameters
        self.material_listbox = tk.Listbox(root, height=5)
        self.material_listbox.grid(row=0, column=1, rowspan=2)
        self.material_listbox.bind(
            '<<ListboxSelect>>', self.on_material_select)

        self.add_material_button = tk.Button(
            root, text='Add Material', command=self.add_material)
        self.add_material_button.grid(row=0, column=2)

        self.remove_material_button = tk.Button(
            root, text='Remove Material', command=self.remove_material)
        self.remove_material_button.grid(row=1, column=2)

        # Parameter fields
        self.fields = {}
        for i, field in enumerate(
                ['Material Name', 'Space Group',
                 'a', 'b', 'c', 'alpha', 'beta', 'gamma']):
            if i > 1:
                units = 'Angstroms' if i < 5 else 'degrees'
                text = f'{field} ({units})'
            else:
                text = field
            tk.Label(root, text=text).grid(row=i+2, column=1)
            entry = tk.Entry(root)
            entry.grid(row=i+2, column=2)
            self.fields[field] = entry

        self.update_button = tk.Button(
            root, text='Update Material Properties',
            command=self.update_material)
        self.update_button.grid(row=11, column=1, columnspan=2)
        self.update_button = tk.Button(
            root, text='Update Material Properties',
            command=self.update_material)
        self.update_button.grid(row=11, column=1, columnspan=2)

        self.confirm_button = tk.Button(
            root, text='Confirm\nAll\nSelected\nMaterial\nProperties',
            command=self.on_close)
        self.confirm_button.grid(row=0, column=3, rowspan=12)

        # Initial Material Data
        if not preselected_materials:
            self.materials = [None]
            self.add_material(None)
        else:
            for material in preselected_materials:
                self.add_material(material)
        self.selected_material = 0

        # Overwrite the root window's close action to call `self.on_close`
        self.root.protocol('WM_DELETE_WINDOW', self.on_close)

    def plot_reference_data(self):
        if self.ref_data_y is None:
            return
        handle = self.ax.plot(
            self.ref_data_x, self.ref_data_y, label=self.ref_data_label)
        self.legend_handles = handle
        self.ax.legend(handles=self.legend_handles)
        self.ax.set_xlabel('Energy (keV)')
        self.ax.set_ylabel('Intensity (a.u)')
        self.canvas.draw()

    def add_material(self, new_material=None):
        # Local modules
        from CHAP.edd.models import MaterialConfig

        if new_material is None:
            new_material = MaterialConfig(
                material_name='Ti64',
                sgnum=194,
                lattice_parameters=[2.9217, 4.66027]
            )
        self.materials.append(new_material)
        self.material_listbox.insert(tk.END, new_material.material_name)
        self.material_listbox.select_set(tk.END)
        self.on_material_select(None)
        self.update_plot()

    def remove_material(self):
        if self.selected_material is not None:
            self.materials.pop(self.selected_material)
            self.material_listbox.delete(self.selected_material)
            self.selected_material = None
            self.clear_fields()
            self.update_plot()

    def update_material(self):
        # Local modules
        from CHAP.edd.utils import make_material
#        from CHAP.utils.material import Material

        if self.selected_material is None:
            return

        material = self.materials[self.selected_material]
        try:
            # Retrieve values from fields
            name = self.fields['Material Name'].get()
            sgnum = int(self.fields['Space Group'].get())
            lattice_parameters = [
                float(self.fields[param].get())
                for param in ('a', 'b', 'c', 'alpha', 'beta', 'gamma')
            ]
            # Make a hexrd material from those values so we can
            # propagate any other updates required by the material's
            # symmetries
            _material = make_material(name, sgnum, lattice_parameters)
#            _material = Material.make_material(
#                name, sgnum=sgnum,
#                lattice_parameters_angstroms=lattice_parameters,
#                pos=['4a', '8c'])
#                #pos=[(0,0,0), (1/4, 1/4, 1/4), (3/4, 3/4, 3/4)])
            lattice_parameters_angstroms = lattice_parameters
            material.material_name = name
            material.sgnum = _material.sgnum
            material.lattice_parameters = [
                _material.latticeParameters[i].value for i in range(6)
            ]
            material._material = _material
            # If the updated field forces other field(s) to get new
            # values (because of space group symmetries), propagate
            # those new values to the gui entries too.
            for key, entry in self.fields.items():
                if key == 'Material Name':
                    continue
                entry.delete(0, tk.END)
            self.fields['Space Group'].insert(0, str(material.sgnum))
            for i, key in enumerate(('a', 'b', 'c', 'alpha', 'beta', 'gamma')):
                self.fields[key].insert(
                    0, str(_material.latticeParameters[i].value))

            # Update the listbox name display
            self.material_listbox.delete(self.selected_material)
            self.material_listbox.insert(
                self.selected_material, material.material_name)
            self.update_plot()
        except ValueError:
            messagebox.showerror(
                'Invalid input',
                'Please enter valid numbers for lattice parameters.')

    def on_material_select(self, event):
        if len(self.material_listbox.curselection()) == 0:
            # Listbox item deselection event can be ignored
            return
        # Update the selected material index
        self.selected_material = self.material_listbox.curselection()[0]
        material = self.materials[self.selected_material]
        self.clear_fields()
        self.fields['Material Name'].insert(0, material.material_name)
        self.fields['Space Group'].insert(0, str(material.sgnum))
        for i, key in enumerate(('a', 'b', 'c', 'alpha', 'beta', 'gamma')):
            self.fields[key].insert(
                0, str(material._material.latticeParameters[i].value))

    def clear_fields(self):
        for entry in self.fields.values():
            entry.delete(0, tk.END)

    def update_plot(self):
        # Local modules
        from CHAP.edd.utils import (
            get_unique_hkls_ds,
            get_peak_locations,
        )

        self.ax.cla()
        self.legend_handles = []
        self.plot_reference_data()  # Re-plot reference data

        # Plot each material's hkl peak locations
        for i, material in enumerate(self.materials):
            hkls, ds = get_unique_hkls_ds([material])
            E0s = get_peak_locations(ds, self.tth)
            for hkl, E0 in zip(hkls, E0s):
                if E0 < min(self.ref_data_x) or E0 > max(self.ref_data_x):
                    continue
                line = self.ax.axvline(
                    E0, c=f'C{i+1}', ls='--', lw=1,
                    label=material.material_name)
                self.ax.text(E0, 1, str(hkl)[1:-1], c=f'C{i+1}',
                             ha='right', va='top', rotation=90,
                             transform=self.ax.get_xaxis_transform())
            self.legend_handles.append(line)
        self.ax.legend(handles=self.legend_handles)
        self.canvas.draw()

    def on_close(self):
        """Handle closing the GUI and triggering the on_complete
        callback."""
        if self.on_complete:
            self.on_complete(self.materials, self.figure)
        self.root.destroy()  # Close the tkinter root window
        plt.close()

def run_material_selector(
        x, y, tth, preselected_materials=None, label='Reference Data',
        on_complete=None, interactive=False):
    """Run the MaterialParamSelector tkinter application.

    :param x: MCA channel energies.
    :type x: np.ndarray
    :param y: MCA intensities.
    :type y: np.ndarray
    :param tth: The (calibrated) 2&theta; angle.
    :type tth: float
    :param preselected_materials: Materials to get HKLs and lattice
        spacings for.
    :type preselected_materials: list[hexrd.material.Material], optional
    :param label: Legend label for the 1D plot of reference MCA data
        from the parameters `x`, `y`, defaults to `"Reference Data"`.
    :type label: str, optional
    :param on_complete: Callback function to handle completion of the
        material selection, defaults to `None`.
    :type on_complete: Callable, optional
    :param interactive: Show the plot and allow user interactions with
        the matplotlib figure, defaults to `False`.
    :type interactive: bool, optional
    :return: The selected materials for the strain analyses.
    :rtype: list[CHAP.edd.models.MaterialConfig]
    """
    # Initialize the main application window
    root = tk.Tk()

    # Create the material parameter selection GUI within the main
    # window
    # This GUI allows the user to adjust and visualize lattice
    # parameters and space group
    app = MaterialParamSelector(
        root, x, y, tth, preselected_materials, label, on_complete)

    if interactive:
        # If interactive mode is enabled, start the GUI event loop to
        # allow user interaction
        root.mainloop()
    else:
        # If not in interactive mode, immediately close the
        # application
        app.on_close()

def select_material_params(
        x, y, tth, preselected_materials=None, label='Reference Data',
        interactive=False, return_buf=False):
    """Interactively adjust the lattice parameters and space group for
    a list of materials. It is possible to add / remove materials from
    the list.

    :param x: MCA channel energies.
    :type x: np.ndarray
    :param y: MCA intensities.
    :type y: np.ndarray
    :param tth: The (calibrated) 2&theta angle.
    :type tth: float
    :param preselected_materials: Materials to get HKLs and
        lattice spacings for.
    :type preselected_materials: list[hexrd.material.Material],
        optional
    :param label: Legend label for the 1D plot of reference MCA data
        from the parameters `x`, `y`, defaults to `"Reference Data"`.
    :type label: str, optional
    :param interactive: Show the plot and allow user interactions with
        the matplotlib figure, defaults to `False`.
    :type interactive: bool, optional
        :param return_buf: Return an in-memory object as a byte stream
        represention of the Matplotlib figure, defaults to `False`.
    :type return_buf: bool, optional
    :return: The selected materials for the strain analyses and a byte
        stream represention of the Matplotlib figure if return_buf is
        `True` (`None` otherwise).
    :rtype: list[CHAP.edd.models.MaterialConfig],
        Union[io.BytesIO, None]
    """
    # Local modules
    from CHAP.utils.general import fig_to_iobuf

    # Run the MaterialParamSelector with the callback function to
    # handle the materials data and, if requested, the output figure
    materials = None
    figure = None
    def on_complete(_materials, _figure):
        nonlocal materials, figure
        materials = _materials
        figure = _figure

    run_material_selector(
        x, y, tth, preselected_materials, label, on_complete, interactive)

    if return_buf:
        buf = fig_to_iobuf(figure)
    else:
        buf = None

    return materials, buf


if __name__ == '__main__':
    # Local modules
    from CHAP.edd.models import MaterialConfig

    xx = np.linspace(40, 100, 100)
    yy = np.sin(xx)
    ttth = 5
    ppreselected_materials = [
        MaterialConfig(
            material_name='Ti64_orig',
            sgnum=194,
            lattice_parameters=[2.9217, 4.66027]
        )
    ]
    mmaterials = select_material_params(
        xx, yy, ttth, preselected_materials=ppreselected_materials,
        interactive=True)
    print(f'Returned materials: {mmaterials}')



#!/usr/bin/env python
"""
File       : reader.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: generic Reader module

Define a generic `Reader` object.
"""

# System modules
import argparse
import logging
import os
from sys import modules

# Third party modules
from pydantic import (
    PrivateAttr,
    constr,
    model_validator,
)

# Local modules
from CHAP.pipeline import PipelineItem


def validate_reader_model(reader):
    reader._mapping_filename = reader.filename
    filename = os.path.normpath(os.path.realpath(
        os.path.join(reader.inputdir, reader.filename)))
    if (not os.path.isfile(filename)
            and not os.path.dirname(reader.filename)):
        reader.logger.warning(
            f'Unable to find {reader.filename} in {reader.inputdir}, looking '
            f'in {reader.outputdir}')
        filename = os.path.normpath(os.path.realpath(
            os.path.join(reader.outputdir, reader.filename)))
    # Note that reader.filename has str type instead of FilePath
    # since its existence is not yet gueranteed (it can be writen
    # over the course of the pipeline's execution). So postpone
    # validation until the entire pipeline gets validated.
    if not os.path.isfile(filename):
        reader.logger.warning(
            f'Unable to find {reader.filename} during validation')
    reader.filename = filename
    return reader


class Reader(PipelineItem):
    """Generic file reader.

    The job of any `Reader` in a `Pipeline` is to provide data stored
    in a file to the next `PipelineItem`. Note that a `Reader` used on
    its own disrupts the flow of data in a `Pipeline` -- it does not
    receive or pass along any data returned by the previous
    `PipelineItem`.

    :ivar filename: Name of file to read from.
    :type filename: str
    """
    filename: constr(strip_whitespace=True, min_length=1)

    _mapping_filename: PrivateAttr(default=None)

    _validate_filename = model_validator(mode='after')(validate_reader_model)

    def read(self):
        """Read and return the contents of `filename` as text.

        :return: The file content.
        :rtype: str
        """
        if not self.filename:
            self.logger.warning(
                'No file name is given, skipping read operation')
            return None
        try:
            with open(self.filename) as f:
                data = f.read()
        except Exception:
            return None
        return data


class OptionParser():
    """User based option parser."""
    def __init__(self):
        self.parser = argparse.ArgumentParser(prog='PROG')
        self.parser.add_argument(
            '--filename', action='store',
            dest='filename', default='', help='Input file')
        self.parser.add_argument(
            '--reader', action='store',
            dest='reader', default='Reader', help='Reader class name')
        self.parser.add_argument(
            '--log-level', choices=logging._nameToLevel.keys(),
            dest='log_level', default='INFO', help='logging level')


def main(opt_parser=OptionParser):
    """Main function."""
    optmgr = opt_parser()
    opts = optmgr.parser.parse_args()
    cls_name = opts.reader
    try:
        reader_cls = getattr(modules[__name__], cls_name)
    except AttributeError:
        print(f'Unsupported reader {cls_name}')
        raise

    reader = reader_cls()
    reader.logger.setLevel(getattr(logging, opts.log_level))
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(logging.Formatter(
        '{name:20}: {message}', style='{'))
    reader.logger.addHandler(log_handler)
    data = reader.read(filename=opts.filename)
    print(f'Reader {reader} reads from {opts.filename}, data {data}')


if __name__ == '__main__':
    main()



"""This subpackage contains `PipelineItems` unique to EDD data
processing workflows.
"""

from CHAP.edd.processor import (
    DiffractionVolumeLengthProcessor,
    LatticeParameterRefinementProcessor,
    HKLProcessor,
    MCAEnergyCalibrationProcessor,
    MCATthCalibrationProcessor,
    ReducedDataProcessor,
    StrainAnalysisProcessor,
)
from CHAP.edd.reader import (
    EddMapReader,
    EddMPIMapReader,
    ScanToMapReader,
    SetupNXdataReader,
    UpdateNXdataReader,
    NXdataSliceReader,
    SliceNXdataReader,
)
from CHAP.edd.writer import (
    StrainAnalysisUpdateWriter,
)



"""
File       : runner.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description:
"""

def parser():
    """Return an argument parser for the `CHAP` CLI. This parser has
    one argument: the input CHAP configuration file.
    """
    # System modules
    from argparse import ArgumentParser
    pparser = ArgumentParser(prog='PROG')

    pparser.add_argument(
        'config', action='store', default='', help='Input configuration file')
    pparser.add_argument(
        '-p', '--pipeline', nargs='*', help='Pipeline name(s)')
    return pparser

def main():
    """Main function."""
    # System modules
    from yaml import safe_load

    # Local modules
    from CHAP.models import RunConfig

    try:
        # Third party modules
        # pylint: disable=c-extension-no-member
        from mpi4py import MPI

        have_mpi = True
        comm = MPI.COMM_WORLD
    except ImportError:
        have_mpi = False
        comm = None

    args = parser().parse_args()

    # Read the input config file
    configfile = args.config
    with open(configfile) as file:
        config = safe_load(file)
    #RV Add to input_files in provenance data writer

    # Check if executed as a worker spawned by another Processor
    run_config = RunConfig(**config.pop('config'), comm=comm)
    if have_mpi and run_config.spawn:
        # pylint: disable=c-extension-no-member
        sub_comm = MPI.Comm.Get_parent()
        common_comm = sub_comm.Merge(True)
        # Read worker specific input config file
        if run_config.spawn > 0:
            with open(f'{configfile}_{common_comm.Get_rank()}') as file:
                config = safe_load(file)
                run_config = RunConfig(
                    **config.pop('config'), comm=common_comm)
        else:
            with open(f'{configfile}_{sub_comm.Get_rank()}') as file:
                config = safe_load(file)
                run_config = RunConfig(**config.pop('config'), comm=comm)
    else:
        common_comm = comm

    # Get the pipeline configurations
    sub_pipelines = args.pipeline
    pipeline_config = []
    if sub_pipelines is None:
        for sub_pipeline in config.values():
            pipeline_config += sub_pipeline
    else:
        for sub_pipeline in sub_pipelines:
            if sub_pipeline in config:
                pipeline_config += config.get(sub_pipeline)
            else:
                raise ValueError(
                    f'Invalid pipeline option: \'{sub_pipeline}\' missing in '
                    f'the pipeline configuration ({list(config.keys())})')

    # Run the pipeline with or without profiling
    if run_config.profile:
        # System modules
        from cProfile import runctx  # python profiler
        from pstats import Stats     # profiler statistics

        cmd = 'runner(run_config, pipeline_config, common_comm)'
        runctx(cmd, globals(), locals(), 'profile.dat')
        info = Stats('profile.dat')
        info.sort_stats('cumulative')
        info.print_stats()
    else:
        runner(run_config, pipeline_config, common_comm)

    # Disconnect the spawned worker
    if have_mpi and run_config.spawn:
        common_comm.barrier()
        sub_comm.Disconnect()

def runner(run_config, pipeline_config, comm=None):
    """Main runner funtion.

    :param run_config: CHAP run configuration.
    :type run_config: CHAP.runner.RunConfig
    :param pipeline_config: CHAP Pipeline configuration.
    :type pipeline_config: dict
    :param comm: MPI communicator.
    :type comm: mpi4py.MPI.Comm, optional
    :return: The pipeline's returned data field.
    """
    # System modules
    from time import time

    # Logging setup
    logger, log_handler = set_logger(run_config.log_level)
    logger.info(f'Input pipeline configuration: {pipeline_config}\n')

    # Run the pipeline
    t0 = time()
    data = run(run_config, pipeline_config, logger, log_handler, comm)
    logger.info(f'Executed "run" in {time()-t0:.3f} seconds')

    return data

def set_logger(log_level='INFO'):
    """Helper function to set CHAP logger.

    :param log_level: Logger level, defaults to `"INFO"`.
    :type log_level: str
    :return: The CHAP logger and logging handler.
    :rtype: logging.Logger, logging.StreamHandler
    """
    # System modules
    import logging

    logger = logging.getLogger(__name__)
    log_level = getattr(logging, log_level.upper())
    logger.setLevel(log_level)
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(logging.Formatter(
        '{asctime}: {name:20}: {levelname}: {message}',
        datefmt='%Y-%m-%d %H:%M:%S', style='{'))
    logger.addHandler(log_handler)
    return logger, log_handler

def run(
        run_config, pipeline_config, logger=None, log_handler=None, comm=None):
    """Run a given pipeline_config.

    :param run_config: CHAP run configuration.
    :type run_config: CHAP.runner.RunConfig
    :param pipeline_config: CHAP Pipeline configuration.
    :type pipeline_config: dict
    :param logger: CHAP logger.
    :type logger: logging.Logger, optional
    :param log_handler: Logging handler.
    :type log_handler: logging.StreamHandler, optional
    :param comm: MPI communicator.
    :type comm: mpi4py.MPI.Comm, optional
    :return: The `data` field of the first item in the returned
        list of pipeline items.
    """
    # System modules
    from logging import getLogger
    import os
    from tempfile import NamedTemporaryFile

    # Local modules
    from CHAP.pipeline import Pipeline

    # Make sure os.makedirs is only called from the root node
    if comm is None:
        rank = 0
    else:
        rank = comm.Get_rank()

    pipeline_args = []
    pipeline_mmcs = []
    for item in pipeline_config:

        # Load individual object with given name from its module
        config = run_config.model_dump()
        if isinstance(item, dict):
            name = list(item.keys())[0]
            item_args = item.get(name)
            if item_args is None:
                item_args = {}
            # Picking "inputdir" and "outputdir" from the item or from
            # the default run configuration, giving precedence to the
            # former
            if 'inputdir' in item_args:
                inputdir = item_args.pop('inputdir')
                if not os.path.isabs(inputdir):
                    inputdir = os.path.normpath(os.path.realpath(
                        os.path.join(run_config.inputdir, inputdir)))
                if not os.path.isdir(inputdir):
                    raise OSError(
                        f'input directory does not exist ({inputdir})')
                if not os.access(inputdir, os.R_OK):
                    raise OSError('input directory is not accessible for '
                                  f'reading ({inputdir})')
                config['inputdir'] = inputdir
            if 'outputdir' in item_args:
                outputdir = item_args.pop('outputdir')
                if not os.path.isabs(outputdir):
                    outputdir = os.path.normpath(os.path.realpath(
                        os.path.join(run_config.outputdir, outputdir)))
                if not rank:
                    if not os.path.isdir(outputdir):
                        os.makedirs(outputdir)
                    try:
                        NamedTemporaryFile(dir=outputdir)
                    except Exception as exc:
                        raise OSError(
                            'output directory is not accessible for '
                            f'writing ({outputdir})') from exc
                config['outputdir'] = outputdir
        else:
            name = item

        # Initialize the object's identifiers
        if 'users' in name:
            # Load users module. This is required in CHAPaaS which can
            # have common area for users module. Otherwise, we will be
            # required to have invidual user's PYTHONPATHs to load user
            # processors.
            try:
                # Third party modules
                # pylint: disable=unused-import
                import users
            except ImportError:
                if logger is not None:
                    logger.error(f'Unable to load {name}')
                continue
            cls_name = name.split('.')[-1]
            mod_name = '.'.join(name.split('.')[:-1])
            module = __import__(mod_name, fromlist=[cls_name])
        else:
            mod_name, cls_name = name.split('.')
            module = __import__(f'CHAP.{mod_name}', fromlist=[cls_name])

        pipeline_mmcs.append(getattr(module, cls_name))

        # Initialize the object's runtime arguments
        item_args['comm'] = comm  #FIX make comm a field in RunConfig?
        if 'name' not in item_args:
            item_args['name'] = cls_name
        item_args.update(config)
        item_logger = getLogger(name)
        if log_handler is not None:
            item_logger.addHandler(log_handler)
        item_args['logger'] = item_logger
        if logger is not None:
            logger.info(
                f'Initialized input fields for an instance of {cls_name}')
        pipeline_args.append(item_args)
    pipeline = Pipeline(mmcs=pipeline_mmcs, args=pipeline_args)
    pipeline.logger.setLevel(run_config.log_level)
    if log_handler is not None:
        pipeline.logger.addHandler(log_handler)
    if logger is not None:
        logger.info(f'Loaded {pipeline} with {len(pipeline_mmcs)} items\n')

    # Make sure os.makedirs completes before continuing all nodes
    if comm is not None:
        comm.barrier()

    # Execute the pipeline
    if logger is not None:
        logger.info(f'Calling "execute" on {pipeline}')
    result = pipeline.execute()
    if result:
        return result[0]['data']
    return result


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""Tomography command line writer."""

# Local modules
from CHAP import Writer


class TomoWriter(Writer):
    """Writer for saving tomo data."""
    def write(self, data):
        """Write the results of the (partial) tomographic
        reconstruction and add provenance data to the data pipeline.

        :param data: Input data.
        :type data: list[PipelineData]
        :return: Output data.
        :rtype: list[PipelineData]
        """
        # System modules
        from os import path as os_path

        # Local modules
        from CHAP.pipeline import PipelineData

        # Load the (partial) tomographic reconstruction result
        tomodata = self.get_data(data, schema='tomodata', remove=self.remove)

        # Local modules
        if isinstance(tomodata, dict):
            from CHAP.common.writer import YAMLWriter as writer
        else:
            from CHAP.common.writer import NexusWriter as writer

        # Write the (partial) tomographic reconstruction result
        #RV FIX make class methods from the Writer.write functions?
        # or create write function that also accept some default type
        # other than list[PipelineData]?
        writer(**self.model_dump()).write(
            [PipelineData(name=self.__name__, data=tomodata)])

        # Add provenance info to the data pipeline
        metadata = self.get_data(data, schema='metadata', remove=False)
        did = metadata['did']
        provenance = {
            'did': did,
            'input_files': [{'name': 'todo.fix: pipeline.yaml'}],
            'output_files': [{'name': os_path.realpath(self.filename)}],
        }
        data.append(PipelineData(
            name=self.__name__, data=provenance, schema='provenance'))

        return data


if __name__ == '__main__':
    # Local modules
    from CHAP.writer import main

    main()



#!/usr/bin/env python
"""EDD specific writers."""

# Local modules
from CHAP import Writer

class StrainAnalysisUpdateWriter(Writer):
    def write(self, data):
        # Third party modules
        from nexusformat.nexus import nxload

        # Local modules
        from CHAP.edd.processor import StrainAnalysisProcessor

        points = self.unwrap_pipelinedata(data)[0]
        nxroot = nxload(self.filename, mode='r+')
        StrainAnalysisProcessor.add_points(nxroot, points, logger=self.logger)

        return nxroot


if __name__ == '__main__':
    # Local modules
    from CHAP.writer import main

    main()



"""Map related Pydantic model classes."""

# System modules
from copy import deepcopy
from functools import (
    cache,
    lru_cache,
)
import os
from typing import (
    Literal,
    Optional,
    Union,
)

# Third party modules
import numpy as np
from pydantic import (
    Field,
    FilePath,
    PrivateAttr,
    conint,
    conlist,
    constr,
    field_validator,
    model_validator,
)
from pyspec.file.spec import FileSpec
from typing_extensions import Annotated

# Local modules
from CHAP.models import CHAPBaseModel


class CHAPSlice(CHAPBaseModel):
    """Class representing a slice configuration for any particular
    dimension of a data set.

    :ivar start: Starting index for slicing, defaults to `0`.
    :type start: int, optional
    :ivar end: Ending index for slicing, defaults to `-1`.
    :type end: int, optional
    :ivar step: Slicing step, defaults to `1`.
    :type step: int, optional
    """
    start: Optional[int] = 0
    end: Optional[int] = None
    step: Optional[conint(gt=0)] = 1

    def tolist(self):
        return [self.start, self.end, self.step]

    def toslice(self):
        return slice(self.start, self.end, self.step)


class Detector(CHAPBaseModel):
    """Class representing a single detector.

    :ivar id: The detector id (e.g. name or channel index).
    :type id: str
    :ivar shape: Shape of detector's raw data.
    :type shape: tuple[int,int]
    :ivar attrs: Additional detector configuration attributes.
    :type attrs: dict, optional
    """
    id_: constr(min_length=1) = Field(alias='id')
    shape: Optional[tuple[int, int]] = None
    attrs: Optional[Annotated[dict, Field(validate_default=True)]] = {}

    @field_validator('id_', mode='before')
    @classmethod
    def validate_id(cls, id_):
        """Validate the detector id.

        :param id: The detector id (e.g. name or channel index).
        :type id: int, str
        :return: The detector id.
        :rtype: str
        """
        if isinstance(id_, int):
            return str(id_)
        return id_

    #RV maybe better to use model_validator, see v2 docs?
    @field_validator('attrs')
    @classmethod
    def validate_attrs(cls, attrs):
        """Validate any additional detector configuration attributes.

        :param attrs: Any additional attributes to `Detector`.
        :type attrs: dict
        :raises ValueError: Invalid attribute.
        :return: The validated field for `attrs`.
        :rtype: dict
        """
        # RV FIX add eta
        name = attrs.get('name')
        if name is not None:
            if isinstance(name, int):
                attrs['name'] = str(name)
            elif not isinstance(name, str):
                raise ValueError
        return attrs

    def get_id(self):
        return self.id_


class DetectorConfig(CHAPBaseModel):
    """Class representing a detector configuration.

    :ivar detectors: Detector list.
    :type detectors: list[Detector]
    :ivar roi: Detector ROI.
    :type roi: list[CHAPSlice, CHAPSlice], optional
    """
    # FIX ROI to make general, now just suited to and tested with TOMO
    detectors: conlist(item_type=Detector)
    roi: Optional[conlist(
        item_type=CHAPSlice, min_length=2, max_length=2)] = None

    @field_validator('roi', mode='before')
    @classmethod
    def validate_roi(cls, roi):
        """Validate the detector ROI.

        :param roi: Detector ROI.
        :type roi: list[CHAPSlice, CHAPSlice]
        :return: The validated detector ROI
        :rtype: list[CHAPSlice, CHAPSlice]
        """
        if roi is None:
            return roi
        return [CHAPSlice().model_dump() if v is None else v for v in roi]


class Sample(CHAPBaseModel):
    """Class representing a sample metadata configuration.

    :ivar name: The name of the sample.
    :type name: str
    :ivar description: A description of the sample.
    :type description: str, optional
    """
    name: constr(min_length=1)
    description: Optional[str] = ''


class SpecScans(CHAPBaseModel):
    """Class representing a set of scans from a single SPEC file.

    :ivar spec_file: Path to the SPEC file.
    :type spec_file: str
    :ivar scan_numbers: List of scan numbers to use.
    :type scan_numbers: Union(int, list[int], str)
    :ivar par_file: Path to a non-default SMB par file.
    :type par_file: str, optional
    """
    spec_file: FilePath
    scan_numbers: Union[
        constr(min_length=1), conlist(item_type=conint(gt=0), min_length=1)]
    par_file: Optional[FilePath] = None

    @field_validator('spec_file')
    @classmethod
    def validate_spec_file(cls, spec_file):
        """Validate the specified SPEC file.

        :param spec_file: Path to the SPEC file.
        :type spec_file: str
        :raises ValueError: If the SPEC file is invalid.
        :return: Absolute path to the SPEC file.
        :rtype: str
        """
        try:
            spec_file = os.path.abspath(spec_file)
            FileSpec(spec_file)
        except Exception as exc:
            raise ValueError(f'Invalid SPEC file {spec_file}') from exc
        return spec_file

    @field_validator('scan_numbers', mode='before')
    @classmethod
    def validate_scan_numbers(cls, scan_numbers, info):
        """Validate the specified list of scan numbers.

        :param scan_numbers: List of scan numbers.
        :type scan_numbers: Union(int, list[int], str)
        :param info: Pydantic validator info object.
        :type info: pydantic_core._pydantic_core.ValidationInfo
        :raises ValueError: If a specified scan number is not found in
            the SPEC file.
        :return: List of scan numbers.
        :rtype: list[int]
        """
        if isinstance(scan_numbers, int):
            scan_numbers = [scan_numbers]
        elif isinstance(scan_numbers, str):
            # Local modules
            from CHAP.utils.general import string_to_list

            scan_numbers = string_to_list(scan_numbers)

        spec_file = info.data.get('spec_file')
        if spec_file is not None:
            spec_scans = FileSpec(spec_file)
            for scan_number in scan_numbers:
                scan = spec_scans.get_scan_by_number(scan_number)
                if scan is None:
                    raise ValueError(
                        f'No scan number {scan_number} in {spec_file}')
        return scan_numbers

    @field_validator('par_file', mode='before')
    @classmethod
    def validate_par_file(cls, par_file):
        """Validate the specified SMB par file.

        :param par_file: Path to a non-default SMB par file.
        :type par_file: str
        :raises ValueError: If the SMB par file is invalid.
        :return: Absolute path to the SMB par file.
        :rtype: str
        """
        if par_file is None or not par_file:
            return None
        par_file = os.path.abspath(par_file)
        if not os.path.isfile(par_file):
            raise ValueError(f'Invalid SMB par file {par_file}')
        return par_file

    @property
    def scanparsers(self):
        """Returns the list of `ScanParser`s for each of the scans
        specified by the SPEC file and scan numbers belonging to this
        instance of `SpecScans`.
        """
        return [self.get_scanparser(scan_no) for scan_no in self.scan_numbers]

    def get_scanparser(self, scan_number):
        """Return a `ScanParser` for the specified scan number in the
        specified SPEC file.

        :param scan_number: Scan number to get a `ScanParser` for.
        :type scan_number: int
        :return: `ScanParser` for the specified scan number.
        :rtype: ScanParser
        """
        if self.par_file is None:
            return get_scanparser(self.spec_file, scan_number)
        return get_scanparser(
            self.spec_file, scan_number, par_file=self.par_file)

    def get_index(self, scan_number, scan_step_index, map_config):
        """Return a tuple representing the index of a specific step in
        a specific SPEC scan within a map.

        :param scan_number: Scan number to get index for.
        :type scan_number: int
        :param scan_step_index: Scan step index to get index for.
        :type scan_step_index: int
        :param map_config: Map configuration to get index for.
        :type map_config: MapConfig
        :return: Index for the specified scan number and scan step
            index within the specified map configuration.
        :rtype: tuple
        """
        index = ()
        for independent_dimension in map_config.independent_dimensions:
            coordinate_index = list(
                map_config.coords[independent_dimension.label]).index(
                    independent_dimension.get_value(
                        self, scan_number, scan_step_index,
                        map_config.scalar_data))
            index = (coordinate_index, *index)
        return index

    def get_detector_data(self, detectors, scan_number, scan_step_index):
        """Return the raw data from the specified detectors at the
        specified scan number and scan step index.

        :param detectors: List of detector prefixes to get raw data
            for.
        :type detectors: list[str]
        :param scan_number: Scan number to get data for.
        :type scan_number: int
        :param scan_step_index: Scan step index to get data for.
        :type scan_step_index: int
        :return: Data from the specified detectors for the specified
            scan number and scan step index.
        :rtype: list[np.ndarray]
        """
        return get_detector_data(
            tuple([detector.prefix for detector in detectors]),
            self.spec_file,
            scan_number,
            scan_step_index)


@cache
def get_available_scan_numbers(spec_file):
    """Get the available scan numbers.

    :param spec_file: Path to the SPEC file.
    :type spec_file: str
    """
    return list(FileSpec(spec_file).scans.keys())


@cache
def get_scanparser(spec_file, scan_number, par_file=None):
    """Get the scanparser.

    :param spec_file: Path to the SPEC file.
    :type spec_file: str
    :param scan_number: Scan number to get data for.
    :type scan_number: int
    :param par_file: Path to a SMB par file.
    :type par_file: str, optional
    """
    # pylint: disable=undefined-variable
    if scan_number not in get_available_scan_numbers(spec_file):
        return None
    if par_file is None:
        return ScanParser(spec_file, scan_number)
    return ScanParser(spec_file, scan_number, par_file=par_file)


@lru_cache(maxsize=10)
def get_detector_data(
        detector_prefixes, spec_file, scan_number, scan_step_index):
    """Get the detector data.

    :param detector_prefixes: The detector prefixes.
    :type detector_prefixes: Union[tuple[str], list[str]]
    :ivar spec_file: Path to the SPEC file.
    :type spec_file: str
    :param scan_number: Scan number to get data for.
    :type scan_number: int
    :param scan_step_index: The scan step index.
    :type scan_step_index: int
    """
    detector_data = []
    scanparser = get_scanparser(spec_file, scan_number)
    for prefix in detector_prefixes:
        image_data = scanparser.get_detector_data(prefix, scan_step_index)
        detector_data.append(image_data)
    return detector_data


class PointByPointScanData(CHAPBaseModel):
    """Class representing a source of raw scalar-valued data for which
    a value was recorded at every point in a `MapConfig`.

    :ivar label: A user-defined label for referring to this data in
        the NeXus file and in other tools.
    :type label: str
    :ivar units: The units in which the data were recorded.
    :type units: str
    :ivar data_type: Represents how these data were recorded at time
        of data collection.
    :type data_type: Literal['spec_motor', 'spec_motor_absolute',
        'scan_column', 'smb_par', 'expression', 'detector_log_timestamps']
    :ivar name: Represents the name with which these raw data were
        recorded at time of data collection.
    :type name: str
    """
    label: constr(min_length=1)
    units: constr(strip_whitespace=True, min_length=1)
    data_type: Literal[
        'spec_motor', 'spec_motor_absolute', 'scan_column', 'smb_par',
        'expression', 'detector_log_timestamps']
    name: constr(strip_whitespace=True, min_length=1)
    ndigits: Optional[conint(ge=0)] = None

    @field_validator('label')
    @classmethod
    def validate_label(cls, label):
        """Validate that the supplied `label` does not conflict with
        any of the values for `label` reserved for certain data needed
        to perform corrections.

        :param label: The value of `label` to validate.
        :type label: str
        :raises ValueError: If `label` is one of the reserved values.
        :return: The originally supplied value `label`.
        :rtype: str
        """
        if ((not issubclass(cls,CorrectionsData))
                and label in CorrectionsData.reserved_labels()):
            raise ValueError(
                f'{cls.__class__.__name__}.label may not be any of the following '
                f'reserved values: {CorrectionsData.reserved_labels()}')
        return label

    def validate_for_station(self, station):
        """Validate this instance of `PointByPointScanData` for a
        certain choice of station (beamline).

        :param station: The name of the station (in 'idxx' format).
        :type station: str
        :raises TypeError: If the station is not compatible with the
            value of the `data_type` attribute for this instance of
            PointByPointScanData.
        """
        if (station.lower() not in ('id1a3', 'id3a')
                and self.data_type == 'smb_par'):
            raise TypeError(
                f'{self.__class__.__name__}.data_type may not be "smb_par" '
                f'when station is "{station}"')
        if (not station.lower() == 'id3b'
                and self.data_type == 'detector_log_timestamps'):
            raise TypeError(
                f'{self.__class__.__name__}.data_type may not be'
                + f' "detector_log_timestamps" when station is "{station}"')

    def validate_for_spec_scans(
            self, spec_scans, scan_step_index='all'):
        """Validate this instance of `PointByPointScanData` for a list
        of `SpecScans`.

        :param spec_scans: A list of `SpecScans` whose raw data will
            be checked for the presence of the data represented by
            this instance of `PointByPointScanData`.
        :type spec_scans: list[SpecScans]
        :param scan_step_index: A specific scan step index to validate,
            defaults to `'all'`.
        :type scan_step_index: Union[Literal['all'],int], optional
        :raises RuntimeError: If the data represented by this instance
            of `PointByPointScanData` is missing for the specified
            scan steps.
        """
        for scans in spec_scans:
            for scan_number in scans.scan_numbers:
                scanparser = scans.get_scanparser(scan_number)
                if scan_step_index == 'all':
                    scan_step_index_range = range(scanparser.spec_scan_npts)
                else:
                    scan_step_index_range = range(
                        scan_step_index, 1+scan_step_index)
                for index in scan_step_index_range:
                    try:
                        self.get_value(scans, scan_number, index)
                    except Exception as exc:
                        raise RuntimeError(
                            f'Could not find data for {self.name} '
                            f'(data_type "{self.data_type}") '
                            f'on scan number {scan_number} '
                            f'for index {index} '
                            f'in spec file {scans.spec_file}') from exc

    def validate_for_scalar_data(self, scalar_data):
        """Used for `PointByPointScanData` objects with a `data_type`
        of `'expression'`. Validate that the `scalar_data` field of a
        `MapConfig` object contains all the items necessary for
        evaluating the expression.

        :param scalar_data: the `scalar_data` field of a `MapConfig`
            that this `PointByPointScanData` object will be validated
            against.
        :type scalar_data: list[PointByPointScanData]
        :raises ValueError: if `scalar_data` does not contain items
           needed for evaluating the expression.
        """
        # Third party modules
        from ast import parse
        from asteval import get_ast_names

        labels = get_ast_names(parse(self.name))
        for label in ('round', 'np', 'numpy'):
            try:
                labels.remove(label)
            except Exception:
                pass
        for l in labels:
            if l in ('round', 'np', 'numpy'):
                continue
            label_found = False
            for s_d in scalar_data:
                if s_d.label == l:
                    label_found = True
                    break
            if not label_found:
                raise ValueError(
                    f'{l} is not the label of an item in scalar_data')

    def get_value(
            self, spec_scans, scan_number, scan_step_index=0,
            scalar_data=None, relative=True, ndigits=None):
        """Return the value recorded for this instance of
        `PointByPointScanData` at a specific scan step.

        :param spec_scans: An instance of `SpecScans` in which the
            requested scan step occurs.
        :type spec_scans: SpecScans
        :param scan_number: The number of the scan in which the
            requested scan step occurs.
        :type scan_number: int
        :param scan_step_index: The index of the requested scan step,
            defaults to `0`.
        :type scan_step_index: int, optional
        :param scalar_data: list of scalar data configurations used to
            get values for `PointByPointScanData` objects with
            `data_type == 'expression'`.
        :type scalar_data: list[PointByPointScanData], optional
        :param relative: Whether to return a relative value or not,
            defaults to `True` (only applies to SPEC motor values).
        :type relative: bool, optional
        :params ndigits: Round SPEC motor values to the specified
            number of decimals if set.
        :type ndigits: int, optional
        :return: The value recorded of the data represented by this
            instance of `PointByPointScanData` at the scan step
            requested.
        :rtype: float
        """
        if 'spec_motor' in self.data_type:
            if ndigits is None:
                ndigits = self.ndigits
            if 'absolute' in self.data_type:
                relative = False
            return get_spec_motor_value(
                spec_scans.spec_file, scan_number, scan_step_index, self.name,
                relative, ndigits)
        if self.data_type == 'scan_column':
            return get_spec_counter_value(
                spec_scans.spec_file, scan_number, scan_step_index, self.name)
        if self.data_type == 'smb_par':
            return get_smb_par_value(
                spec_scans.spec_file, scan_number, self.name)
        if self.data_type == 'expression':
            if scalar_data is None:
                scalar_data = []
            return get_expression_value(
                spec_scans, scan_number, scan_step_index, self.name,
                scalar_data)
        if self.data_type == 'detector_log_timestamps':
            timestamps = get_detector_log_timestamps(
                spec_scans.spec_file, scan_number, self.name)
            if scan_step_index >= 0:
                return timestamps[scan_step_index]
            return timestamps
        return None

@cache
def get_spec_motor_value(
        spec_file, scan_number, scan_step_index, spec_mnemonic,
        relative=True, ndigits=None):
    """Return the value recorded for a SPEC motor at a specific scan
    step.

    :param spec_file: Location of a SPEC file in which the requested
        scan step occurs.
    :type spec_scans: str
    :param scan_number: The number of the scan in which the requested
        scan step occurs.
    :type scan_number: int
    :param scan_step_index: The index of the requested scan step.
    :type scan_step_index: int
    :param spec_mnemonic: The menmonic of a SPEC motor.
    :type spec_mnemonic: str
    :param relative: Whether to return a relative value or not,
        defaults to `True`.
    :type relative: bool, optional
    :params ndigits: Round SPEC motor values to the specified
        number of decimals if set.
    :type ndigits: int, optional
    :return: The value of the motor at the scan step requested.
    :rtype: float
    """
    scanparser = get_scanparser(spec_file, scan_number)
    if (hasattr(scanparser, 'spec_scan_motor_mnes')
            and spec_mnemonic in scanparser.spec_scan_motor_mnes):
        motor_i = scanparser.spec_scan_motor_mnes.index(spec_mnemonic)
        if scan_step_index >= 0:
            scan_step = np.unravel_index(
                scan_step_index,
                scanparser.spec_scan_shape,
                order='F')
            motor_value = \
                scanparser.get_spec_scan_motor_vals(
                    relative)[motor_i][scan_step[motor_i]]
        else:
            motor_value = scanparser.get_spec_scan_motor_vals(
                relative)[motor_i]
            if len(scanparser.spec_scan_shape) == 2:
                if motor_i == 0:
                    motor_value = np.concatenate(
                        [motor_value] * scanparser.spec_scan_shape[1])
                else:
                    motor_value = np.repeat(
                        motor_value, scanparser.spec_scan_shape[0])
    else:
        motor_value = scanparser.get_spec_positioner_value(spec_mnemonic)
    if ndigits is not None:
        motor_value = np.round(motor_value, ndigits)
    return motor_value

@cache
def get_spec_counter_value(
        spec_file, scan_number, scan_step_index, spec_column_label):
    """Return the value recorded for a SPEC counter at a specific scan
    step.

    :param spec_file: Location of a SPEC file in which the requested
        scan step occurs.
    :type spec_scans: str
    :param scan_number: The number of the scan in which the requested
        scan step occurs.
    :type scan_number: int
    :param scan_step_index: The index of the requested scan step.
    :type scan_step_index: int
    :param spec_column_label: The label of a SPEC data column.
    :type spec_column_label: str
    :return: The value of the counter at the scan step requested.
    :rtype: float
    """
    scanparser = get_scanparser(spec_file, scan_number)
    if scan_step_index >= 0:
        return scanparser.spec_scan_data[spec_column_label][scan_step_index]
    return scanparser.spec_scan_data[spec_column_label]

@cache
def get_smb_par_value(spec_file, scan_number, par_name):
    """Return the value recorded for a specific scan in SMB-tyle .par
    file.

    :param spec_file: Location of a SPEC file in which the requested
        scan step occurs.
    :type spec_scans: str
    :param scan_number: The number of the scan in which the requested
        scan step occurs.
    :type scan_number: int
    :param par_name: The name of the column in the .par file.
    :type par_name: str
    :return: The value of the .par file value for  the scan requested.
    :rtype: float
    """
    scanparser = get_scanparser(spec_file, scan_number)
    return scanparser.pars[par_name]

def get_expression_value(
        spec_scans, scan_number, scan_step_index, expression,
        scalar_data):
    """Return the value of an evaluated expression of other sources of
    point-by-point scalar scan data for a single point.

    :param spec_scans: An instance of `SpecScans` in which the
        requested scan step occurs.
    :type spec_scans: SpecScans
    :param scan_number: The number of the scan in which the requested
        scan step occurs.
    :type scan_number: int
    :param scan_step_index: The index of the requested scan step.
    :type scan_step_index: int
    :param expression: The string expression to evaluate.
    :type expression: str
    :param scalar_data: the `scalar_data` field of a `MapConfig`
        object (used to provide values for variables used in
        `expression`).
    :type scalar_data: list[PointByPointScanData]
    :return: The value of the .par file value for  the scan requested.
    :rtype: float
    """
    # Third party modules
    from ast import parse
    from asteval import get_ast_names, Interpreter

    labels = get_ast_names(parse(expression))
    symtable = {}
    for l in labels:
        if l == 'round':
            symtable[l] = round
        for s_d in scalar_data:
            if s_d.label == l:
                symtable[l] = s_d.get_value(
                    spec_scans, scan_number, scan_step_index, scalar_data)
    aeval = Interpreter(symtable=symtable)
    return aeval(expression)

@cache
def get_detector_log_timestamps(spec_file, scan_number, detector_prefix):
    """Return the list of detector timestamps for the given scan &
    detector prefix.

    :param spec_file: Location of a SPEC file in which the requested
        scan occurs.
    :type spec_scans: str
    :param scan_number: The number of the scan for which to return
        detector log timestamps.
    :type scan_number: int
    :param detector_prefix: The prefix of the detecotr whose log file
        should be used.
    :return: All detector log timestamps for the given scan.
    :rtype: list[float]
    """
    sp = get_scanparser(spec_file, scan_number)
    return sp.get_detector_log_timestamps(detector_prefix)

def validate_data_source_for_map_config(data_source, info):
    """Confirm that an instance of PointByPointScanData is valid for
    the station and scans provided by a map configuration dictionary.

    :param data_source: The input object to validate.
    :type data_source: PointByPointScanData
    :param info: Pydantic validator info object.
    :type info: pydantic_core._pydantic_core.ValidationInfo
    :raises Exception: If `data_source` cannot be validated.
    :return: The validated `data_source` instance.
    :rtype: PointByPointScanData
    """
    def _validate_data_source_for_map_config(data_source, info):
        if isinstance(data_source, list):
            return [_validate_data_source_for_map_config(d_s, info)
                    for d_s in data_source]
        if data_source is not None:
            values = info.data
            if data_source.data_type == 'expression':
                data_source.validate_for_scalar_data(values['scalar_data'])
            else:
                import_scanparser(
                    values['station'], values['experiment_type'])
                data_source.validate_for_station(values['station'])
                if values['validate_data_present']:
                    data_source.validate_for_spec_scans(values['spec_scans'])
        return data_source

    return _validate_data_source_for_map_config(data_source, info)


class IndependentDimension(PointByPointScanData):
    """Class representing the source of data to identify the
    coordinate values along one dimension of a `MapConfig`.

    :ivar start: Sarting index for slicing all datasets of a
        `MapConfig` along this axis, defaults to `0`.
    :type start: int, optional
    :ivar end: Ending index for slicing all datasets of a `MapConfig`
        along this axis, defaults to the total number of unique values
        along this axis in the associated `MapConfig`.
    :type end: int, optional
    :ivar step: Step for slicing all datasets of a `MapConfig` along
        this axis, defaults to `1`.
    :type step: int, optional
    """
    # FIX convert to using CHAPSlice
    start: Optional[conint(ge=0)] = 0
    end: Optional[int] = None
    step: Optional[conint(gt=0)] = 1

#    @field_validator('step')
#    @classmethod
#    def validate_step(cls, step):
#        """Validate that the supplied value of `step`.
#
#        :param step: The value of `step` to validate.
#        :type step: str
#        :raises ValueError: If `step` is zero.
#        :return: The originally supplied value `step`.
#        :rtype: int
#        """
#        if step == 0 :
#            raise ValueError('slice step cannot be zero')
#        return step


class CorrectionsData(PointByPointScanData):
    """Class representing the special instances of
    `PointByPointScanData` that are used by certain kinds of
    `CorrectionConfig` tools.

    :ivar label: One of the reserved values required by
        `CorrectionConfig`.
    :type label: Literal['presample_intensity', 'postsample_intensity',
                         'dwell_time_actual']
    :ivar data_type: Represents how these data were recorded at time
        of data collection.
    :type data_type: Literal['scan_column', 'smb_par']
    """
    label: Literal['presample_intensity', 'postsample_intensity',
                   'dwell_time_actual']
    data_type: Literal['scan_column','smb_par']

    @classmethod
    def reserved_labels(cls):
        """Return a list of all the labels reserved for
        corrections-related scalar data.

        :return: A list of reserved labels.
        :rtype: list[str]
        """
        return list((*cls.model_fields['label'].annotation.__args__, 'round'))


class PresampleIntensity(CorrectionsData):
    """Class representing a source of raw data for the intensity of
    the beam that is incident on the sample.

    :ivar label: Must be `'presample_intensity"`.
    :type label: Literal['presample_intensity']
    :ivar units: Must be `'counts'`.
    :type units: Literal['counts']
    """
    label: Literal['presample_intensity'] = 'presample_intensity'
    units: Literal['counts'] = 'counts'


class PostsampleIntensity(CorrectionsData):
    """Class representing a source of raw data for the intensity of
    the beam that has passed through the sample.

    :ivar label: Must be `'postsample_intensity'`.
    :type label: Literal['postsample_intensity']
    :ivar units: Must be `'counts'`.
    :type units: Literal['counts']
    """
    label: Literal['postsample_intensity'] = 'postsample_intensity'
    units: Literal['counts'] = 'counts'


class DwellTimeActual(CorrectionsData):
    """Class representing a source of raw data for the actual dwell
    time at each scan point in SPEC (with some scan types, this value
    can vary slightly point-to-point from the dwell time specified in
    the command).

    :ivar label: Must be `'dwell_time_actual'`.
    :type label: Literal['dwell_time_actual']
    :ivar units: Must be `'s'`.
    :type units: Literal['s']
    """
    label: Literal['dwell_time_actual'] = 'dwell_time_actual'
    units: Literal['s'] = 's'


class SpecConfig(CHAPBaseModel):
    """Class representing the raw data for one or more SPEC scans.

    :ivar station: The name of the station at which the data was
        collected.
    :type station: Literal['id1a3', 'id3a', 'id3b', 'id4b']
    :ivar spec_scans: A list of the SPEC scans that compose the set.
    :type spec_scans: list[SpecScans]
    :ivar experiment_type: Experiment type.
    :type experiment_type: Literal['EDD', 'GIWAXS', 'SAXSWAXS', 'TOMO',
        'XRF', 'HDRM']
    """
    station: Literal['id1a3', 'id3a', 'id3b', 'id4b']
    experiment_type: Literal[
        'EDD', 'GIWAXS', 'SAXSWAXS', 'TOMO', 'XRF', 'HDRM']
    spec_scans: conlist(item_type=SpecScans, min_length=1)

    @model_validator(mode='before')
    @classmethod
    def validate_specconfig_before(cls, data):
        """Ensure that a valid configuration was provided and finalize
        spec_file filepaths.

        :param data: Pydantic validator data object.
        :type data: SpecConfig,
            pydantic_core._pydantic_core.ValidationInfo
        :return: The currently validated list of class properties.
        :rtype: dict
        """
        inputdir = data.get('inputdir')
        if inputdir is not None:
            spec_scans = data.get('spec_scans')
            for i, scans in enumerate(deepcopy(spec_scans)):
                if isinstance(scans, dict):
                    spec_file = scans['spec_file']
                    if not os.path.isabs(spec_file):
                        spec_scans[i]['spec_file'] = os.path.join(
                            inputdir, spec_file)
                else:
                    spec_file = scans.spec_file
                    if not os.path.isabs(spec_file):
                        spec_scans[i].spec_file = os.path.join(
                            inputdir, spec_file)
            data['spec_scans'] = spec_scans
        return data

    @field_validator('experiment_type')
    @classmethod
    def validate_experiment_type(cls, experiment_type, info):
        """Ensure values for the station and experiment_type fields are
        compatible.

        :param experiment_type: The value of `experiment_type` to
            validate.
        :type experiment_type: str
        :param info: Pydantic validator info object.
        :type info: pydantic_core._pydantic_core.ValidationInfo
        :raises ValueError: Invalid experiment type.
        :return: The validated field for `experiment_type`.
        :rtype: str
        """
        station = info.data.get('station')
        if station == 'id1a3':
            allowed_experiment_types = ['EDD', 'SAXSWAXS', 'TOMO']
        elif station == 'id3a':
            allowed_experiment_types = ['EDD', 'TOMO']
        elif station == 'id3b':
            allowed_experiment_types = ['GIWAXS', 'SAXSWAXS', 'TOMO', 'XRF']
        elif station == 'id4b':
            allowed_experiment_types = ['HDRM']
        else:
            allowed_experiment_types = []
        if experiment_type not in allowed_experiment_types:
            raise ValueError(
                f'For station {station}, allowed experiment types are '
                f'{", ".join(allowed_experiment_types)}. '
                f'Supplied experiment type {experiment_type} is not allowed.')
        import_scanparser(station, experiment_type)
        return experiment_type


class MapConfig(CHAPBaseModel):
    """Class representing an experiment consisting of one or more SPEC
    scans.

    :param: did: FOXDEN data identifier.
    :type did: str, optional
    :ivar title: The title for the map configuration.
    :type title: str
    :ivar station: The name of the station at which the map was
        collected.
    :type station: Literal['id1a3', 'id3a', 'id3b', id4b]
    :ivar experiment_type: Experiment type.
    :type experiment_type: Literal['EDD', 'GIWAXS', 'SAXSWAXS', 'TOMO',
        'XRF', 'HDRM']
    :ivar sample: The sample metadata configuration.
    :type sample: CHAP.commom.models.map.Sample
    :ivar spec_scans: A list of the SPEC scans that compose the map.
    :type spec_scans: list[SpecScans]
    :ivar scalar_data: A list of the sources of data representing
        other scalar raw data values collected at each point on the
        map. In the NeXus file representation of the map, datasets for
        these values will be included, defaults to `[]`.
    :type scalar_data: list[PointByPointScanData], optional
    :ivar independent_dimensions: A list of the sources of data
        representing the raw values of each independent dimension of
        the map.
    :type independent_dimensions: list[PointByPointScanData]
    :ivar presample_intensity: A source of point-by-point presample
        beam intensity data. Required when applying a CorrectionConfig
        tool.
    :type presample_intensity: PresampleIntensity, optional
    :ivar dwell_time_actual: A source of point-by-point actual dwell
        times for SPEC scans. Required when applying a
        CorrectionConfig tool.
    :type dwell_time_actual: DwellTimeActual, optional
    :ivar postsample_intensity: A source of point-by-point postsample
        beam intensity data. Required when applying a CorrectionConfig
        tool with `correction_type='flux_absorption'` or
        `correction_type='flux_absorption_background'`.
    :type postsample_intensity: PresampleIntensity, optional
    :ivar attrs: Additional Map configuration attributes.
    :type attrs: dict, optional
    """
    validate_data_present: bool = True
    did: Optional[constr(strip_whitespace=True)] = None
    title: constr(strip_whitespace=True, min_length=1)
    station: Literal['id1a3', 'id3a', 'id3b', 'id4b']
    experiment_type: Literal[
        'EDD', 'GIWAXS', 'SAXSWAXS', 'TOMO', 'XRF', 'HDRM']
    sample: Sample
    spec_scans: conlist(item_type=SpecScans, min_length=1)
    scalar_data: Optional[conlist(item_type=PointByPointScanData)] = []
    independent_dimensions: conlist(
        item_type=IndependentDimension, min_length=1)
    presample_intensity: Optional[PresampleIntensity] = None
    dwell_time_actual: Optional[DwellTimeActual] = None
    postsample_intensity: Optional[PostsampleIntensity] = None
    attrs: dict = {}
#    _coords: dict = PrivateAttr()
    _dims: tuple = PrivateAttr()
#    _scan_step_indices: list = PrivateAttr()
#    _shape: tuple = PrivateAttr()

    _validate_independent_dimensions = field_validator(
        'independent_dimensions')(validate_data_source_for_map_config)
    _validate_presample_intensity = field_validator(
        'presample_intensity')(validate_data_source_for_map_config)
    _validate_dwell_time_actual = field_validator(
        'dwell_time_actual')(validate_data_source_for_map_config)
    _validate_postsample_intensity = field_validator(
        'postsample_intensity')(validate_data_source_for_map_config)
    _validate_scalar_data = field_validator(
        'scalar_data')(validate_data_source_for_map_config)

    @model_validator(mode='before')
    @classmethod
    def validate_mapconfig_before(cls, data, info):
        """Ensure that a valid configuration was provided and finalize
        spec_file filepaths.

        :param data: Pydantic validator data object.
        :type data:
            MapConfig, pydantic_core._pydantic_core.ValidationInfo
        :return: The currently validated list of class properties.
        :rtype: dict
        """
        if 'spec_file' in data and 'scan_numbers' in data:
            spec_file = data.pop('spec_file')
            scan_numbers = data.pop('scan_numbers')
            if 'par_file' in data:
                par_file = data.pop('par_file')
            else:
                par_file = None
            if 'spec_scans' in data:
                raise ValueError(
                    f'Ambiguous SPEC scan information: spec_file={spec_file},'
                    f' scan_numbers={scan_numbers}, and '
                    f'spec_scans={data["spec_scans"]}')
            if par_file is None:
                data['spec_scans'] = [
                    {'spec_file': spec_file, 'scan_numbers': scan_numbers}]
            else:
                data['spec_scans'] = [
                    {'spec_file': spec_file, 'scan_numbers': scan_numbers,
                     'par_file': par_file}]
        else:
            spec_scans = data.get('spec_scans')
            if 'spec_scans' in data:
                inputdir = data.get('inputdir')
                if inputdir is None and info.data is not None:
                    inputdir = info.data.get('inputdir')
                for i, scans in enumerate(deepcopy(spec_scans)):
                    if isinstance(scans, SpecScans):
                        scans = scans.model_dump()
                    spec_file = scans['spec_file']
                    if inputdir is not None and not os.path.isabs(spec_file):
                        scans['spec_file'] = os.path.join(inputdir, spec_file)
                    spec_scans[i] = SpecScans(**scans, **data)
            data['spec_scans'] = spec_scans
        return data

    @field_validator('experiment_type')
    @classmethod
    def validate_experiment_type(cls, experiment_type, info):
        """Ensure values for the station and experiment_type fields are
        compatible.

        :param experiment_type: The value of `experiment_type` to
            validate.
        :type experiment_type: dict
        :param info: Pydantic validator info object.
        :type info: pydantic_core._pydantic_core.ValidationInfo
        :raises ValueError: Invalid experiment type.
        :return: The validated field for `experiment_type`.
        :rtype: str
        """
        station = info.data['station']
        if station == 'id1a3':
            allowed_experiment_types = ['EDD', 'SAXSWAXS', 'TOMO']
        elif station == 'id3a':
            allowed_experiment_types = ['EDD', 'TOMO']
        elif station == 'id3b':
            allowed_experiment_types = ['GIWAXS', 'SAXSWAXS', 'TOMO', 'XRF']
        elif station == 'id4b':
            allowed_experiment_types = ['HDRM']
        else:
            allowed_experiment_types = []
        if experiment_type not in allowed_experiment_types:
            raise ValueError(
                f'For station {station}, allowed experiment types are '
                f'{", ".join(allowed_experiment_types)}. '
                f'Supplied experiment type {experiment_type} is not allowed.')
        return experiment_type


    @model_validator(mode='before')
    @classmethod
    def validate_before(cls, data):
        if data.get('attrs') is None:
            data['attrs'] = {}
        return data

    #RV maybe better to use model_validator, see v2 docs?
    @field_validator('attrs')
    @classmethod
    def validate_attrs(cls, attrs, info):
        """Validate any additional attributes depending on the values
        for the station and experiment_type fields.

        :param attrs: Any additional attributes to the MapConfig class.
        :type attrs: dict
        :param info: Pydantic validator info object.
        :type info: pydantic_core._pydantic_core.ValidationInfo
        :raises ValueError: Invalid attribute.
        :return: The validated field for `attrs`.
        :rtype: dict
        """
        # Get the map's scan_type for EDD experiments
        values = info.data
        station = values['station']
        experiment_type = values['experiment_type']
        if station in ['id1a3', 'id3a'] and experiment_type == 'EDD':
            scan_type = cls.get_smb_par_attr(values, 'scan_type')
            if scan_type is not None:
                attrs['scan_type'] = scan_type
            attrs['config_id'] = cls.get_smb_par_attr(values, 'config_id')
            dataset_id = cls.get_smb_par_attr(
                values, 'dataset_id', unique=False)
            if dataset_id is not None:
                attrs['dataset_id'] = dataset_id
            if attrs.get('scan_type') is None:
                return attrs
            axes_labels = {1: 'fly_labx', 2: 'fly_laby', 3: 'fly_labz',
                           4: 'fly_ometotal'}
            if attrs['scan_type'] != 0:
                attrs['fly_axis_labels'] = [
                    axes_labels[cls.get_smb_par_attr(values, 'fly_axis0')]]
            if attrs['scan_type'] in (2, 3, 5):
                attrs['fly_axis_labels'].append(
                    axes_labels[cls.get_smb_par_attr(values, 'fly_axis1')])
        return attrs

    @staticmethod
    def get_smb_par_attr(
        class_fields, label, units='-', name=None, unique=True):
        """Read an SMB par file attribute.

        :param class_fields: The Map configuration class fields.
        :type class_fields: Any
        :param label: An attrs field key, the user-defined label for
            referring to this data in the NeXus file and in other
            tools.
        :type label: str
        :param units: The attrs' field unit, defaults to `'-'`.
        :type units: str
        :param name: The attrs' field name, the name with which these
            raw data were recorded at time of data collection,
            defaults to `label`.
        :type name: str, optional.
        """
        if name is None:
            name = label
        PointByPointScanData(
            label=label, data_type='smb_par', units=units, name=name)
        values = []
        for scans in class_fields.get('spec_scans'):
            for scan_number in scans.scan_numbers:
                scanparser = scans.get_scanparser(scan_number)
                try:
                    values.append(scanparser.pars[name])
                except Exception:
#                    print(
#                        f'Warning: No value found for .par file value "{name}"'
#                        f' on scan {scan_number} in spec file '
#                        f'{scans.spec_file}.')
                    values.append(None)
        values = list(set(values))
        if len(values) == 1:
            return values[0]
        if unique:
            raise ValueError(
                f'More than one {name} in map not allowed ({values})')
        return values

    @property
    def all_scalar_data(self):
        """Return a list of all instances of `PointByPointScanData`
        for which this map configuration will collect dataset-like
        data (as opposed to axes-like data).

        This will be any and all of the items in the
        corrections-data-related fields, as well as any additional
        items in the optional `scalar_data` field.
        """
        return [getattr(self, label, None)
                for label in CorrectionsData.reserved_labels()
                if getattr(self, label, None) is not None] + self.scalar_data

    @property
    def coords(self):
        """Return a dictionary of the values of each independent
        dimension across the map.
        """
        raise RuntimeError('property coords not implemented')
        if not hasattr(self, '_coords'):
            fly_axis_labels = self.attrs.get('fly_axis_labels', [])
            coords = {}
            for dim in self.independent_dimensions:
                if dim.label in fly_axis_labels:
                    relative = True
                    ndigits = 3
                else:
                    relative = False
                    ndigits = None
                coords[dim.label] = []
                for scans in self.spec_scans:
                    for scan_number in scans.scan_numbers:
                        scanparser = scans.get_scanparser(scan_number)
                        for scan_step_index in range(
                                scanparser.spec_scan_npts):
                            coords[dim.label].append(dim.get_value(
                                    scans, scan_number, scan_step_index,
                                    self.scalar_data, relative, ndigits))
                if self.map_type == 'structured':
                    coords[dim.label] = np.unique(coords[dim.label])
            self._coords = coords
        return self._coords

    @property
    def dims(self):
        """Return a tuple of the independent dimension labels for the
        map.
        """
        if not hasattr(self, '_dims'):
            self._dims = [dim.label for dim in self.independent_dimensions]
        return self._dims

    @property
    def scan_step_indices(self):
        """Return an ordered list in which we can look up the SpecScans
        object, the scan number, and scan step index for every point
        on the map.
        """
        raise RuntimeError('property scan_step_indices not implemented')
        if not hasattr(self, '_scan_step_indices'):
            scan_step_indices = []
            for scans in self.spec_scans:
                for scan_number in scans.scan_numbers:
                    scanparser = scans.get_scanparser(scan_number)
                    for scan_step_index in range(scanparser.spec_scan_npts):
                        scan_step_indices.append(
                            (scans, scan_number, scan_step_index))
            self._scan_step_indices = scan_step_indices
        return self._scan_step_indices

    @property
    def shape(self):
        """Return the shape of the map -- a tuple representing the
        number of unique values of each dimension across the map.
        """
        raise RuntimeError('property shape not implemented')
        if not hasattr(self, '_shape'):
            if self.map_type == 'structured':
                self._shape = tuple([len(v) for k, v in self.coords.items()])
            else:
                self._shape =  (len(self.scan_step_indices),)
        return self._shape

    def get_coords(self, map_index):
        """Return a dictionary of the coordinate names and values of
        each independent dimension for a given point on the map.

        :param map_index: The map index to return coordinates for.
        :type map_index: tuple
        :return: A list of coordinate values.
        :rtype: dict
        """
        raise RuntimeError('get_coords not implemented')
        if self.map_type == 'structured':
            scan_type = self.attrs.get('scan_type', -1)
            fly_axis_labels = self.attrs.get('fly_axis_labels', [])
            if (scan_type in (3, 5)
                    and len(self.dims) ==
                        len(map_index) + len(fly_axis_labels)):
                dims = [dim for dim in self.dims if dim not in fly_axis_labels]
                return {dim:self.coords[dim][i]
                        for dim, i in zip(dims, map_index)}
            return {dim:self.coords[dim][i]
                    for dim, i in zip(self.dims, map_index)}
        return {dim:self.coords[dim][map_index[0]] for dim in self.dims}

    def get_detector_data(self, detector_name, map_index):
        """Return detector data collected by this map for a given
        point on the map.

        :param detector_name: Name of the detector for which to return
            data. Usually the value of the detector's EPICS
            areaDetector prefix macro, $P.
        :type detector_name: str
        :param map_index: The map index to return detector data for.
        :type map_index: tuple
        :return: One frame of raw detector data.
        :rtype: np.ndarray
        """
        raise RuntimeError('get_detector_data not implemented')
        scans, scan_number, scan_step_index = \
            self.get_scan_step_index(map_index)
        scanparser = scans.get_scanparser(scan_number)
        return scanparser.get_detector_data(detector_name, scan_step_index)

    def get_scan_step_index(self, map_index):
        """Return parameters to identify a single SPEC scan step that
        corresponds to the map point at the index provided.

        :param map_index: The index of a map point to identify as a
            specific SPEC scan step index.
        :type map_index: tuple
        :return: A `SpecScans` configuration, scan number, and scan
            step index.
        :rtype: tuple[SpecScans, int, int]
        """
        raise RuntimeError('get_scan_step_index not implemented')
        fly_axis_labels = self.attrs.get('fly_axis_labels', [])
        if self.map_type == 'structured':
            map_coords = self.get_coords(map_index)
            for scans, scan_number, scan_step_index in self.scan_step_indices:
                coords = {dim.label:(
                              dim.get_value(
                                  scans, scan_number, scan_step_index,
                                  self.scalar_data, True, 3)
                              if dim.label in fly_axis_labels
                              else
                              dim.get_value(
                                  scans, scan_number, scan_step_index,
                                  self.scalar_data))
                          for dim in self.independent_dimensions}
                if coords == map_coords:
                    return scans, scan_number, scan_step_index
            raise RuntimeError(f'Unable to match coordinates {coords}')
        return self.scan_step_indices[map_index[0]]

    def get_value(self, data, map_index):
        """Return the raw data collected by a single device at a
        single point in the map.

        :param data: The device configuration to return a value of raw
            data for.
        :type data: PointByPointScanData
        :param map_index: The map index to return raw data for.
        :type map_index: tuple
        :return: Raw data value.
        """
        raise RuntimeError('get_value not implemented')
        scans, scan_number, scan_step_index = \
            self.get_scan_step_index(map_index)
        return data.get_value(scans, scan_number, scan_step_index,
                              self.scalar_data)


def import_scanparser(station, experiment):
    """Given the name of a CHESS station and experiment type, import
    the corresponding subclass of `ScanParser` as `ScanParser`.

    :param station: The station name
        ('IDxx', not the beamline acronym).
    :type station: str
    :param experiment: The experiment type.
    :type experiment: Literal[
        'EDD', 'GIWAXS', 'SAXSWAXS', 'TOMO', 'XRF', 'HDRM']
    """
    # Third party modules
    # pylint: disable=import-error
    from chess_scanparsers import choose_scanparser

    globals()['ScanParser'] = choose_scanparser(station, experiment)



"""Functions for converting between some commonly used data
formats."""


def convert_sparse_dense(data):
    """Converts between dense and sparse representations for NumPy
    arrays, xarray DataArrays, and xarray Datasets.

    - If input is a NumPy array, converts to a SciPy sparse CSR matrix.
    - If input is a SciPy sparse matrix, converts to a dense NumPy array.
    - If input is an xarray DataArray or Dataset containing sparse
      arrays, converts to dense.
    - If input is an xarray DataArray or Dataset containing dense
      arrays, converts to sparse.

    :param data: The imput data.
    :type data: Union[numpy.ndarray, scipy.sparse.spmatrix,
        xarray.DataArray, xarray.Dataset]
    :return: Converted object -- sparse if input is dense, dense if
        input is sparse.
    """
    import numpy as np
    import scipy.sparse as sp
    import xarray as xr

    if isinstance(data, np.ndarray):
        return sp.csr_matrix(data)  # Convert dense NumPy array to sparse

    if sp.issparse(data):
        return data.toarray()  # Convert sparse matrix to dense NumPy array

    if isinstance(data, xr.DataArray):
        # Convert DataArray values while preserving metadata
        if sp.issparse(data.data):
            return xr.DataArray(
                data.data.toarray(), dims=data.dims,
                coords=data.coords, attrs=data.attrs)
        return xr.DataArray(
            sp.csr_matrix(data.data), dims=data.dims,
            coords=data.coords, attrs=data.attrs)

    if isinstance(data, xr.Dataset):
        # Convert each variable in the Dataset
        def convert_var(var):
            if sp.issparse(var.data):
                return xr.DataArray(var.data.toarray(), dims=var.dims,
                                    coords=var.coords, attrs=var.attrs)
            return xr.DataArray(sp.csr_matrix(var.data), dims=var.dims,
                                coords=var.coords, attrs=var.attrs)

        return data.map(convert_var)

    raise TypeError(f'Unsupported data type: {type(data)}. '
                    'Input must be a NumPy array, SciPy sparse matrix, '
                    'xarray DataArray, or xarray Dataset.')


def convert_xarray_nexus(data):
    """Convert an `xarray.DataArray` or `xarray.Dataset` into an
    `nexusformat.nexus.NXdata` or vice versa.

    :param data: Input data.
    :type data: Union[xarray.DataArray, xarray.Dataset,
        nexusformat.nexus.NXdata]
    :return: Conveted data.
    :rtype:  Union[xarray.DataArray, xarray.Dataset,
        nexusformat.nexus.NXdata]
    """
    import xarray as xr
    from nexusformat.nexus import NXdata, NXfield

    if isinstance(data, xr.DataArray):
        return NXdata(
            value=data.values,
            name=data.name,
            attrs=data.attrs,
            axes=tuple(
                NXfield(
                    value=data[dim].values,
                    name=dim,
                    attrs=data[dim].attrs,
                )
                for dim in data.dims),
        )
    if isinstance(data, xr.Dataset):
        return NXdata(
            **{var:
               NXfield(
                   value=data[var].values,
                   name=var,
                   attrs=data[var].attrs,
               )
               for var in data.data_vars},
            name=data.name,
            attrs=data.attrs,
            axes=tuple(
                NXfield(
                    value=data[dim].values,
                    name=dim,
                    attrs=data[dim].attrs,
                )
                for dim in data.dims),
        )
    if isinstance(data, NXdata):
        nxaxes = data.nxaxes
        if nxaxes is None:
            if 'unstructured_axes' in data.attrs:
                nxaxes = data.unstructured_axes
        if isinstance(nxaxes, str):
            nxaxes = [nxaxes]
        return xr.Dataset(
            data_vars={
                name: xr.DataArray(
                    data=field.nxdata,
                    name=name,
                    attrs=field.attrs,
                )
                for name, field in data.items()
                if isinstance(field, NXfield)
                and name not in nxaxes
            },
            coords={
                axis: xr.DataArray(
                    data=data[axis].nxdata,
                    name=data[axis].nxname,
                    attrs=data[axis].attrs,
                )
                for axis in nxaxes
            },
            attrs=data.attrs,
        )
    raise TypeError(f'Unsupported data type: {type(data)}. '
                    'Must be xarray.DataArray, xarray.Dataset, or NXdata.')


def convert_structured_unstructured(data):
    from copy import deepcopy
    from nexusformat.nexus import NXdata, NXfield
    import numpy as np

    if isinstance(data, NXdata):
        if 'unstructured_axes' in data.attrs:
            # Convert unstructured to structured
            nxaxes = data.attrs['unstructured_axes']
            attrs = deepcopy(data.attrs)
            attrs['axes'] = attrs['unstructured_axes']
            attrs.pop('unstructured_axes')
            signals = [name for name, child in data.items()
                       if name not in nxaxes]
            structured_axes = {a: np.unique(data[a].nxdata) for a in nxaxes}
            dataset_shape = tuple(len(v) for a, v in structured_axes.items())
            structured_signals = {s: np.empty(
                (*dataset_shape, *data[s].shape[1:]),
                dtype=data[s].nxdata.dtype,
            ) for s in signals}
            npts = len(data[signals[0]].nxdata.tolist())
            indices = {
                a: np.searchsorted(structured_axes[a], data[a].nxdata)
                for a in nxaxes
            }
            for s, value in structured_signals.items():
                value[tuple(indices[a] for a in nxaxes)] = data[s]
            structured_data = NXdata(
                **{s: NXfield(
                    value=structured_signals[s],
                    name=s,
                    attrs=data[s].attrs,
                ) for s in signals},
                name=data.nxname,
                attrs=attrs,
                axes=tuple(
                    NXfield(
                        value=value,
                        name=a,
                        attrs={k: v for k, v in data[a].attrs.items()
                               if not k == 'target'},
                    )
                    for a, value in structured_axes.items()
                )
            )
            return structured_data

        if 'axes' in data.attrs:
            # Convert structued to unstructured
            raise NotImplementedError(
                'Conversion from structured to unstructured not implemented.')
    raise TypeError(f'Unsupported data type: {type(data)}')



#!/usr/bin/env python
"""
File       : reader.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Module for Readers used in multiple experiment-specific workflows.
"""

# System modules
from typing import (
    Literal,
    Optional,
    Union,
)

# Third party modules
import numpy as np
from pydantic import (
    PrivateAttr,
    conint,
    conlist,
    constr,
    model_validator,
)

# Local modules
from CHAP import Reader
from CHAP.reader import validate_reader_model
from CHAP.common.models.map import (
    DetectorConfig,
    SpecConfig,
)

def validate_model(model):
    if model.filename is not None:
        validate_reader_model(model)
    return model


class BinaryFileReader(Reader):
    """Reader for binary files."""
    def read(self):
        """Return a content of a given binary file.

        :return: File content.
        :rtype: binary
        """
        with open(self.filename, 'rb') as file:
            data = file.read()
        return data


class ConfigReader(Reader):
    """Reader for YAML files that optionally implements and verifies it
    agaist its Pydantic configuration schema.
    """
    def read(self):
        """Return an optionally verified dictionary from the contents
        of a yaml file.
        """
        data = YAMLReader(**self.model_dump()).read()
        #print(f'\nConfigReader.read start data {type(data)}:')
        raise RuntimeError(
            'FIX ConfigReader downstream validators do not like a pydantic '
            'class as output of a reader, but returning data.model_dict() '
            'instead screws up default value identification')
        #pprint(data)
        if self.get_schema() is not None:
            data = self.get_config(config=data, schema=self.get_schema())
        self.status = 'read'
        #print(f'\nConfigReader.read end data {type(data)}:')
        #pprint(data)
        return data


class DetectorDataReader(Reader):
    """Reader for detector data files. Glob filenames allowed. Mask
    application and background correction available."""
    def read(self, filename,
             mask_file=None, mask_above=None, mask_below=None,
             mask_value=np.nan,
             data_scalar=None,
             background_file=None, background_scalar=None,
             ):
        """Reads detector data, applies masking, scaling, and
        background subtraction.

        :param filename: Path to the primary data file.
        :type filename: str
        :param mask_file: Path to the mask file (optional).
        :type mask_file: str, optional
        :param mask_above: Mask values above this threshold (optional).
        :type mask_above: float, optional
        :param mask_below: Mask values below this threshold (optional).
        :type mask_below: float, optional
        :param data_scalar: Scalar to multiply the data (optional).
        :type data_scalar: float, optional
        :param background_file: Path to the background file (optional).
        :type background_file: str, optional
        :param background_scalar: Scalar to multiply the background
            data (optional).
        :type background_scalar: float, optional
        :return: Processed detector data.
        :rtype: numpy.ndarray
        """
        import glob
        import os
        import numpy as np

        # Handle glob filenames
        if not os.path.isfile(filename):
            filenames = sorted(glob.glob(filename))
            if not filenames:
                raise ValueError(
                    f'{filename} is not a file or glob that matches any files')
        else:
            filenames = [filename]

        # Read the raw data files
        self.logger.info(f'Reading {len(filenames)} raw data files')
        raw_data = [self.get_data_from_file(f) for f in filenames]

        # Initialize mask arary
        self.logger.info(f'Initializing mask array')
        mask_array = np.zeros_like(raw_data[0], dtype=bool)
        if mask_file:
            mask_array |= self.get_data_from_file(mask_file) != 0

        # Initialize background data
        self.logger.info('Initializing background data')
        if background_file:
            background_data = self.get_data_from_file(background_file)
            if background_scalar:
                background_data *= background_scalar
        else:
            background_data = None

        # Handle mask_value of NaN
        if isinstance(mask_value, str):
            if mask_value.lower() == 'nan':
                mask_value = np.nan

        # Scale data, apply mask, subtract background
        self.logger.info('Applying corrections to raw data')
        corrected_data = self.correct_data(
            np.array(raw_data),
            mask_array, mask_above, mask_below, mask_value,
            data_scalar, background_data)

        return corrected_data

    def get_data_from_file(self, filename):
        import fabio

        self.logger.debug(f'Reading {filename}')
        with fabio.open(filename) as datafile:
            data = datafile.data
        return data

    def correct_data(self, raw_data,
                     mask_array, mask_above, mask_below, mask_value,
                     data_scalar,
                     background_data):
        import numpy as np

        # Scale raw data
        corrected_data = raw_data
        if data_scalar:
            corrected_data *= data_scalar

        # Apply mask
        mask = mask_array != 0
        if mask_above is not None:
            mask |= corrected_data > mask_above
        if mask_below is not None:
            mask |= corrected_data < mask_below
        corrected_data = np.where(mask, mask_value, raw_data)

        # Subtract background
        if background_data:
            corrected_data -= background_data

        return corrected_data


class FabioImageReader(Reader):
    """Reader for images using the python package.
    [`fabio`](https://fabio.readthedocs.io/en/main/).

    :ivar frame: Index of a specific frame to read from the file(s),
        defaults to `None`.
    :type frame: int, optional
    """
    frame: Optional[conint(ge=0)] = None

    def read(self):
        """Return the data from the image file(s) provided.

        :returns: Image data as a numpy array (or list of numpy
            arrays, if a glob pattern matching more than one file was
            provided).
        :rtype: Union[numpy.ndarray, list[numpy.ndarray]]
        """
        # Third party modules
        from glob import glob
        import fabio

        filenames = glob(self.filename)
        data = []
        for f in filenames:
            image = fabio.open(f, frame=self.frame)
            data.append(image.data)
            image.close()
        return data


class H5Reader(Reader):
    """Reader for h5 files.

    :ivar h5path: Path to a specific location in the h5 file to read
        data from, defaults to `'/'`.
    :type h5path: str, optional
    :ivar idx: Data slice to read from the object at the specified
        location in the h5 file.
    :type idx: list[int], optional

    """
    h5path: Optional[constr(strip_whitespace=True, min_length=1)] = '/'
    idx: Optional[conlist(min_length=1, max_length=3, item_type=int)] = None

    def read(self):
        """Return the data object stored at `h5path` in an h5-file.

        :return: Object indicated by `filename` and `h5path`.
        :rtype: object
        """
        # Third party modules
        from h5py import File

        data = File(self.filename, 'r')[self.h5path]
        if self.idx is not None:
            data = data[tuple(self.idx)]
        return data


class LinkamReader(Reader):
    """Reader for loading Linkam load frame .txt files as an
    `NXdata`.

    :ivar columns: Column names to read in, defaults to None
        (read in all columns)
    :type columns: list[str], optional
    """
    columns: Optional[conlist(
        item_type=constr(strip_whitespace=True, min_length=1))] = None

    def read(self):
        """Read specified columns from the given Linkam file.

        :returns: Linkam data represented in an `NXdata` object
        :rtype: nexusformat.nexus.NXdata
        """
        # Third party modules
        from nexusformat.nexus import (
            NXdata,
            NXfield,
        )

        # Parse .txt file
        start_time, metadata, data = LinkamReader.parse_file(
            self.filename, self.logger)

        # Get list of actual data column names and corresponding
        # signal nxnames (same as user-supplied column names)
        signal_names = []
        if self.columns is None:
            signal_names = [(col, col) for col in data.keys() if col != 'Time']
        else:
            for col in self.columns:
                col_actual = col
                if col == 'Distance':
                    col_actual = 'Force V Distance_X'
                elif col == 'Force':
                    col_actual = 'Force V Distance_Y'
                elif not col in data:
                    if f'{col}_Y' in data:
                        # Always use the *_Y column if the user-supplied
                        # column name has both _X and _Y components
                        col_actual = f'{col}_Y'
                    else:
                        self.logger.warning(
                            f'{col} not present in {self.filename}')
                        continue
                signal_names.append((col_actual, col))
        self.logger.info(f'Using (column name, signal name): {signal_names}')

        nxdata = NXdata(
            axes=(NXfield(
                name='Time',
                value=np.array(data['Time']) + start_time,
                dtype='float64',
            ),),
            **{col: NXfield(
                name=col,
                value=data[col_actual],
                dtype='float32',
            ) for col_actual, col in signal_names},
            attrs=metadata
        )
        return nxdata

    @classmethod
    def parse_file(cls, filename, logger):
        """Return start time, metadata, and data stored in the
        provided Linkam .txt file.

        :returns:
        :rtype: tuple(float, dict[str, str], dict[str, list[float]])
        """
        # System modules
        from datetime import datetime
        import os
        import re

        # Get t=0 from filename
        start_time = None
        basename = os.path.basename(filename)
        pattern = r'(\d{2}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}-\d{2})'
        match = re.search(pattern, basename)
        if match:
            datetime_str = match.group(1)
            dt = datetime.strptime(datetime_str, '%d-%m-%y_%H-%M-%S-%f')
            start_time = dt.timestamp()
        else:
            logger.warning(f'Datetime not found in {filename}')

        # Get data add metadata from file contents
        metadata = {}
        data = False
        with open(filename, 'r', encoding='utf-8') as inf:
            for line in inf:
                line = line.strip()
                if not line:
                    continue
                if data:
                    # If data dict has been initialized, remaining
                    # lines are all data values
                    values = line.replace(',', '').split('\t')
                    for val, col in zip(values, list(data.keys())):
                        try:
                            val = float(val)
                        except Exception as exc:
                            logger.warning(
                                f'Cannot convert {col} value to float: {val} '
                                f'({exc})')
                            continue
                        else:
                            data[col].append(val)
                if ':' in line:
                    # Metadata key: value pair kept on this line
                    _metadata = line.split(':', 1)
                    if len(_metadata) == 2:
                        key, val = _metadata
                    else:
                        continue
                        key, val = _metadata[0], None
                    metadata[key] = val
                if re.match(r'^([\w\s\w]+)(\t\t[\w\s\w]+)*$', line):
                    # Match found for start of data section -- this
                    # line and the next are column labels.
                    data_cols = []
                    # Get base quantity column names
                    base_cols = line.split('\t\t')
                    # Get Index, X and Y component columns
                    line = next(inf)
                    comp_cols = line.split('\t')
                    # Assemble actual column names
                    data_cols.append('Index')
                    comp_cols_count = int((len(comp_cols) - 1) / 2)
                    for i in range(comp_cols_count):
                        data_cols.extend(
                            [f'{base_cols[i]}_{comp}' for comp in ('X', 'Y')]
                        )
                    if len(base_cols) > comp_cols_count:
                        data_cols.extend(base_cols[comp_cols_count - 1:])
                    # First column (after 0th) is actually Time
                    data_cols[1] = 'Time'
                    # Start of data lines
                    data = {col: [] for col in data_cols}
                    logger.info(f'Found data columns: {data_cols}')

        return start_time, metadata, data


class MapReader(Reader):
    """Reader for CHESS sample maps."""
    def read(
            self, filename=None, map_config=None, detector_names=None,
            inputdir=None):
        """Take a map configuration dictionary and return a
        representation of the map as a NeXus NXentry object. The
        NXentry's default data group will contain the raw data
        collected over the course of the map.

        :param filename: The name of a file with the map configuration
            to read and pass onto the constructor of
            `CHAP.common.models.map.MapConfig`.
        :type filename: str, optional
        :param map_config: A map configuration to be passed directly to
            the constructor of `CHAP.common.models.map.MapConfig`.
        :type map_config: dict, optional
        :param detector_names: Detector prefixes to include raw data
            for in the returned NeXus NXentry object.
        :type detector_names: list[str], optional
        :param inputdir: Input directory, used only if files in the
            input configuration are not absolute paths,
            defaults to `'.'`.
        :type inputdir: str, optional
        :return: Data from the provided map configuration.
        :rtype: nexusformat.nexus.NXentry
        """
        # Third party modules
        from nexusformat.nexus import (
            NXcollection,
            NXdata,
            NXentry,
            NXfield,
            NXsample,
        )

        # Local modules
        from CHAP.common.models.map import MapConfig

        raise RuntimeError('MapReader is obsolete, use MapProcessor')

        if filename is not None:
            if map_config is not None:
                raise RuntimeError('Specify either filename or map_config '
                                   'in common.MapReader, not both')
            # Read the map configuration from file
            if not isfile(filename):
                raise OSError(f'input file does not exist ({filename})')
            extension = splitext(filename)[1]
            if extension in ('.yml', '.yaml'):
                reader = YAMLReader()
            else:
                raise RuntimeError('input file has a non-implemented '
                                   f'extension ({filename})')
            map_config = reader.read(filename)
        elif not isinstance(map_config, dict):
            raise RuntimeError('Invalid parameter map_config in '
                               f'common.MapReader ({map_config})')

        # Validate the map configuration provided by constructing a
        # MapConfig
        map_config = MapConfig(**map_config, inputdir=inputdir)

        # Set up NXentry and add misc. CHESS-specific metadata
        nxentry = NXentry(name=map_config.title)
        nxentry.attrs['station'] = map_config.station
        nxentry.map_config = map_config.model_dump_json()
        nxentry.spec_scans = NXcollection()
        for scans in map_config.spec_scans:
            nxentry.spec_scans[scans.scanparsers[0].scan_name] = \
                NXfield(value=scans.scan_numbers,
                        attrs={'spec_file': str(scans.spec_file)})

        # Add sample metadata
        nxentry[map_config.sample.name] = NXsample(
            **map_config.sample.model_dump())

        # Set up default data group
        nxentry.data = NXdata()
        if map_config.map_type == 'structured':
            nxentry.data.attrs['axes'] = map_config.dims
        for dim in map_config.independent_dimensions:
            nxentry.data[dim.label] = NXfield(
                value=map_config.coords[dim.label],
                units=dim.units,
                attrs={'long_name': f'{dim.label} ({dim.units})',
                       'data_type': dim.data_type,
                       'local_name': dim.name})

        # Create empty NXfields for all scalar data present in the
        # provided map configuration
        signal = False
        auxilliary_signals = []
        for data in map_config.all_scalar_data:
            nxentry.data[data.label] = NXfield(
                value=np.zeros(map_config.shape),
                units=data.units,
                attrs={'long_name': f'{data.label} ({data.units})',
                       'data_type': data.data_type,
                       'local_name': data.name})
            if not signal:
                signal = data.label
            else:
                auxilliary_signals.append(data.label)
        if signal:
            nxentry.data.attrs['signal'] = signal
            nxentry.data.attrs['auxilliary_signals'] = auxilliary_signals

        # Create empty NXfields of appropriate shape for raw
        # detector data
        if detector_names is None:
            detector_names = []
        for detector_name in detector_names:
            if not isinstance(detector_name, str):
                detector_name = str(detector_name)
            detector_data = map_config.get_detector_data(
                detector_name, (0,) * len(map_config.shape))
            nxentry.data[detector_name] = NXfield(value=np.zeros(
                (*map_config.shape, *detector_data.shape)),
                dtype=detector_data.dtype)

        # Read and fill in maps of raw data
        if len(map_config.all_scalar_data) > 0 or detector_names:
            for map_index in np.ndindex(map_config.shape):
                for data in map_config.all_scalar_data:
                    nxentry.data[data.label][map_index] = map_config.get_value(
                        data, map_index)
                for detector_name in detector_names:
                    if not isinstance(detector_name, str):
                        detector_name = str(detector_name)
                    nxentry.data[detector_name][map_index] = \
                        map_config.get_detector_data(detector_name, map_index)

        return nxentry


class PandasReader(Reader):
    """Reader for files that can be read in with
    (pandas)[https://pandas.pydata.org/docs/index.html]"""
    def read(self, filename, method='read_csv', comment='#', kwargs=None):
        """Return a `pandas.DataFrame` read from the given file.

        :param filename: Name of file to read from.
        :type filename: str
        :param method: Name of `pandas` method to use for reading from
            `filename`. Defaults to `'read_csv'`.
        :type method: str, optional
        :param comment: Character to identify comment lines in the
            input file, defaults to `'#'`.
        :type comment: str, optional
        :param kwargs: Additional keyword arguments to supply to the
            `pandas` reader.
        :param kwargs: dict[str, object], optional.
        :rtype: `pandas.DataFrame`
        """
        # Third party modules
        import pandas as pd

        reader = getattr(pd, method)
        if not callable(reader):
            raise ValueError(
                f'{method} is not a callable pandas reader method')

        if kwargs is None:
            kwargs = {}
        if not isinstance(kwargs, dict):
            raise TypeError(
                f'Invalid kwargs type ({type(kwargs)}, should be dict)')

        return reader(filename, comment=comment, **kwargs)


class NexusReader(Reader):
    """Reader for NeXus files.

    :ivar nxpath: Path to a specific location in the NeXus file tree
        to read from, defaults to `'/'`.
    :type nxpath: str, optional
    :ivar idx: Index of array to select, defaults to `None`
    :type idx: int, optional
    :ivar mode: File mode, defaults to 'r'.
    :type mode: Literal['r', 'rw', 'r+', 'w', 'a'], optional
    :ivar nxmemory: Maximum memory usage when reading NeXus files.
    :type nxmemory: int, optional
    """
    nxpath: Optional[constr(strip_whitespace=True, min_length=1)] = '/'
    idx: Optional[conint(ge=0)] = None
    mode: Literal['r', 'rw', 'r+', 'w', 'a'] = 'r'
    nxmemory: Optional[conint(gt=0)] = None

    def read(self):
        """Return the NeXus object stored at `nxpath` in a NeXus file.

        :raises nexusformat.nexus.NeXusError: If `filename` is not a
            NeXus file or `nxpath` is not in its tree.
        :return: NeXus object indicated by `filename` and `nxpath`.
        :rtype: nexusformat.nexus.NXobject
        """
        # Third party modules
        from nexusformat.nexus import (
            nxload,
            nxsetconfig,
        )

        if self.nxmemory is not None:
            nxsetconfig(memory=self.nxmemory)
        if self.idx is not None:
            return nxload(self.filename, mode=self.mode)[self.nxpath][self.idx]
        return nxload(self.filename, mode=self.mode)[self.nxpath]


class NXdataReader(Reader):
    """Reader for constructing an NXdata object from components."""
    def read(self, name, nxfield_params, signal_name, axes_names, attrs=None):
        """Return a basic NXdata object constructed from components.

        :param name: NXdata group name.
        :type name: str
        :param nxfield_params: List of sets of parameters for
            `NXfieldReader` specifying the NXfields belonging to the
            NXdata.
        :type nxfield_params: list[dict]
        :param signal_name: Name of the signal for the NXdata (must be
            one of the names of the NXfields indicated in `nxfields`).
        :type signal: str
        :param axes_names: Name or names of the coordinate axes
            NXfields associated with the signal (must be names of
            NXfields indicated in `nxfields`).
        :type axes_names: Union[str, list[str]]
        :param attrs: Dictionary of additional attributes for the
            NXdata.
        :type attrs: dict, optional
        :returns: A new NXdata object.
        :rtype: nexusformat.nexus.NXdata
        """
        # Third party modules
        from nexusformat.nexus import NXdata

        # Read in NXfields
        nxfields = [NXfieldReader().read(**params, inputdir=self.inputdir)
                    for params in nxfield_params]
        nxfields = {nxfield.nxname: nxfield for nxfield in nxfields}

        # Get signal NXfield
        try:
            nxsignal = nxfields[signal_name]
        except Exception as exc:
            raise ValueError(
                '`signal_name` must be the name of one of the NXfields '
                'indicated in `nxfields`: , '.join(nxfields.keys())) from exc

        # Get axes NXfield(s)
        if isinstance(axes_names, str):
            axes_names = [axes_names]
        try:
            nxaxes = [nxfields[axis_name] for axis_name in axes_names]
        except Exception as exc:
            raise ValueError(
                '`axes_names` must contain only names of NXfields indicated '
                'in `nxfields`: ' + ', '.join(nxfields.keys())) from exc
        for i, nxaxis in enumerate(nxaxes):
            if len(nxaxis) != nxsignal.shape[i]:
                raise ValueError(
                    f'Shape mismatch on signal dimension {i}: signal '
                    + f'"{nxsignal.nxname}" has {nxsignal.shape[i]} values, '
                    + f'but axis "{nxaxis.nxname}" has {len(nxaxis)} values.')

        if attrs is None:
            attrs = {}
        result = NXdata(signal=nxsignal, axes=nxaxes, name=name, attrs=attrs,
                        **nxfields)
        self.logger.info(result.tree)
        return result


class NXfieldReader(Reader):
    """Reader for an NXfield with options to modify certain attributes.
    """
    def read(self, nxpath, nxname=None, update_attrs=None, slice_params=None):
        """Return a copy of the indicated NXfield from the file. Name
        and attributes of the returned copy may be modified with the
        `nxname` and `update_attrs` keyword arguments.

        :param nxpath: Path in `nxfile` pointing to the NXfield to
           read.
        :type nxpath: str
        :param nxname: New name for the returned NXfield.
        :type nxname: str, optional
        :param update_attrs: Optional dictonary used to add to /
            update the original NXfield's attributes.
        :type update_attrs: dict, optional
        :param slice_params: Parameters for returning just a slice of
            the full field data. Slice parameters are provided in a
            list dictionaries with integer values for any / all of the
            following keys: `"start"`, `"end"`, `"step"`. Default
            values used are: `"start"` - `0`, `"end"` -- `None`,
            `"step"` -- `1`. The order of the list must correspond to
            the order of the field's axes.
        :type slice_params: list[dict[str, int]], optional
        :returns: A copy of the indicated NXfield (with name and
            attributes optionally modified).
        :rtype: nexusformat.nexus.NXfield
        """
        # Third party modules
        from nexusformat.nexus import (
            NXfield,
            nxload,
        )

        nxroot = nxload(self.filename)
        nxfield = nxroot[nxpath]

        if nxname is None:
            nxname = nxfield.nxname

        attrs = nxfield.attrs
        if update_attrs is not None:
            attrs.update(update_attrs)

        if slice_params is None:
            value = nxfield.nxdata
        else:
            if len(slice_params) < nxfield.ndim:
                slice_params.extend([{}] * (nxfield.ndim - len(slice_params)))
            if len(slice_params) > nxfield.ndim:
                slice_params = slice_params[0:nxfield.ndim]
            slices = ()
            # FIX convert to using CHAPSlice
            default_slice = {'start': 0, 'end': None, 'step': 1}
            for s in slice_params:
                for k, v in default_slice.items():
                    if k not in s:
                        s[k] = v
                slices = (*slices, slice(s['start'], s['end'], s['step']))
            value = nxfield.nxdata[slices]

        nxfield = NXfield(value=value, name=nxname, attrs=attrs)
        self.logger.debug(f'Result -- nxfield.tree =\n{nxfield.tree}')

        return nxfield


class SpecReader(Reader):
    """Reader for CHESS SPEC scans.

    :ivar config: SPEC configuration to be passed directly to the
        constructor of `CHAP.common.models.map.SpecConfig`.
    :type config: dict, optional
    :ivar detectors: Detector configurations of the detectors to
        include raw data for in the returned NeXus NXroot object,
        defaults to None (only a valid input for EDD).
    :type detectors: Union[
        dict, common.models.map.DetectorConfig], optional
    :ivar filename: Name of file to read from.
    :type filename: str, optional
    """
    config: Optional[Union[dict, SpecConfig]] = None
    detector_config: Optional[DetectorConfig] = None
    filename: Optional[str] = None

    _mapping_filename: PrivateAttr(default=None)

    _validate_filename = model_validator(mode='after')(validate_model)

    @model_validator(mode='after')
    def validate_specreader_after(self):
        """Validate the `SpecReader` configuration.

        :return: The validated configuration.
        :rtype: PipelineItem
        """
        if self.filename is not None:
            if self.config is not None:
                raise ValueError('Specify either filename or config in '
                       'common.SpecReader, not both')
            self.config = YAMLReader(**self.model_dump()).read()
        self.config = self.get_config(
            config=self.config, schema='common.models.map.SpecConfig')
        if self.detector_config is None:
            if self.config.experiment_type != 'EDD':
                raise RuntimeError(
                    'Missing parameter detector_config for experiment type '
                    f'{self.config.experiment_type}')
        return self

    def read(self):
        """Take a SPEC configuration filename or dictionary and return
        the raw data as a NeXus NXentry object.

        :return: Data from the provided SPEC configuration.
        :rtype: nexusformat.nexus.NXroot
        """
        # Third party modules
        # pylint: disable=no-name-in-module
        from json import dumps
        from nexusformat.nexus import (
            NXcollection,
            NXdata,
            NXentry,
            NXfield,
            NXroot,
        )
        # pylint: enable=no-name-in-module

        # Local modules
        from CHAP.common.models.map import Detector

        # Create the NXroot object
        nxroot = NXroot()
        nxentry = NXentry(name=self.config.experiment_type)
        nxroot[nxentry.nxname] = nxentry

        # Set up NXentry and add misc. CHESS-specific metadata as well
        # as all spec_motors, scan_columns, and smb_pars, and the
        # detector info and raw detector data
        nxentry.config = self.config.model_dump_json()
        nxentry.attrs['station'] = self.config.station
        nxentry.spec_scans = NXcollection()
#        nxpaths = []
        if self.config.experiment_type == 'EDD':
            detector_data_format = None
        for scans in self.config.spec_scans:
            nxscans = NXcollection()
            nxentry.spec_scans[f'{scans.scanparsers[0].scan_name}'] = nxscans
            nxscans.attrs['spec_file'] = str(scans.spec_file)
            nxscans.attrs['scan_numbers'] = scans.scan_numbers
            for scan_number in scans.scan_numbers:
                scanparser = scans.get_scanparser(scan_number)
                if self.config.experiment_type == 'EDD':
                    if detector_data_format is None:
                        detector_data_format = scanparser.detector_data_format
                    elif (scanparser.detector_data_format !=
                            detector_data_format):
                        raise NotImplementedError(
                            'Mixing `spec` and `h5` data formats')
                    if self.detector_config is None:
                        detectors_ids = None
                    elif detector_data_format == 'spec':
                        raise NotImplementedError(
                            'detector_data_format = "spec"')
                    else:
                        detectors_ids = [
                            int(d.get_id())
                            for d in self.detector_config.detectors]
                nxscans[scan_number] = NXcollection()
                try:
                    nxscans[scan_number].spec_motors = dumps(
                        {k:float(v) for k,v
                         in scanparser.spec_positioner_values.items()})
                except Exception:
                    pass
                try:
                    nxscans[scan_number].scan_columns = dumps(
                        {k:list(v) for k,v
                         in scanparser.spec_scan_data.items() if len(v)})
                except Exception:
                    pass
                try:
                    nxscans[scan_number].smb_pars = dumps(
                        {k:v for k,v in scanparser.pars.items()})
                except Exception:
                    pass
                try:
                    nxscans[scan_number].spec_scan_motor_mnes = dumps(
                        scanparser.spec_scan_motor_mnes)
                except Exception:
                    pass
                if self.config.experiment_type == 'EDD':
                    nxdata = NXdata()
                    nxscans[scan_number].data = nxdata
#                    nxpaths.append(
#                        f'spec_scans/{nxscans.nxname}/{scan_number}/data')
                    nxdata.data = NXfield(
                        value=scanparser.get_detector_data(detectors_ids)[0])
                else:
                    if self.config.experiment_type == 'TOMO':
                        dtype = np.float32
                    else:
                        dtype = None
                    nxdata = NXdata()
                    nxscans[scan_number].data = nxdata
#                    nxpaths.append(
#                        f'spec_scans/{nxscans.nxname}/{scan_number}/data')
                    for detector in self.detector_config.detectors:
                        if self.detector_config.roi is None:
                            detector_roi = [None, None]
                        else:
                            detector_roi=[
                               self.detector_config.roi[0].toslice(),
                               self.detector_config.roi[1].toslice()]
                        nxdata[detector.get_id()] = NXfield(
                           value=scanparser.get_detector_data(
                               detector.get_id(),
                               detector_roi=detector_roi,
                               dtype=dtype))

        if (self.config.experiment_type == 'EDD' and
                self.detector_config is None):
            if detector_data_format == 'spec':
                raise NotImplementedError('detector_data_format = "spec"')
            self.detector_config = DetectorConfig(
                detectors=[
                    Detector(id=i) for i in range(nxdata.data.shape[1])])
        nxentry.detectors = self.detector_config.model_dump_json()

        #return nxroot, nxpaths
        return nxroot


class URLReader(Reader):
    """Reader for data available over HTTPS."""
    def read(self, url, headers=None, timeout=10):
        """Make an HTTPS request to the provided URL and return the
        results. Headers for the request are optional.

        :param url: URL to read.
        :type url: str
        :param headers: Headers to attach to the request.
        :type headers: dict, optional
        :param timeout: Timeout for the HTTPS request,
            defaults to `10`.
        :type timeout: int
        :return: Content of the response.
        :rtype: object
        """
        # System modules
        import requests

        if headers is None:
            headers = {}
        resp = requests.get(url, headers=headers, timeout=timeout)
        data = resp.content

        self.logger.debug(f'Response content: {data}')

        return data


class YAMLReader(Reader):
    """Reader for YAML files."""
    def read(self):
        """Return a dictionary from the contents of a yaml file.

        :return: Contents of the file.
        :rtype: dict
        """
        # Third party modules
        import yaml

        with open(self.filename) as f:
            data = yaml.safe_load(f)
        return data

class ZarrReader(Reader):
    """Reader for Zarr stores."""

    def read(self, filename, path='/', idx=None, mode='r'):
        """Return the Zarr object stored at `path` in a Zarr store.

        :param filename: Path or URL to the Zarr store.
        :type filename: str
        :param path: Path to a specific location in the Zarr hierarchy
            to read from, defaults to `'/'`.
        :type path: str, optional
        :param idx: Optional index or slice to apply to the returned object.
        :type idx: int, slice, tuple, optional
        :param mode: Store access mode, defaults to `'r'`.
            Common values: `'r'`, `'r+'`, `'a'`, `'w'`.
        :type mode: str, optional
        :raises KeyError: If `path` does not exist in the store.
        :return: The Zarr array or group indicated by `filename` and `path`.
        :rtype: zarr.core.Array or zarr.hierarchy.Group
        """
        import zarr

        # Open the Zarr store (directory, zip, or URL)
        root = zarr.open(filename, mode=mode)

        # Normalize path handling
        if path == '/' or path == '':
            data = root
        else:
            # Remove leading slash for Zarr traversal
            data = root[path.lstrip('/')]

        # Optional indexing (arrays only)
        if idx is not None:
            data = data[idx]

        return data


if __name__ == '__main__':
    # Local modules
    from CHAP.reader import main

    main()



#!/usr/bin/env python
#-*- coding: utf-8 -*-
"""
File       : server.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Python server with thread pool and CHAP pipeline

===========
Client side
===========
cat /tmp/chap.json
{"pipeline": [{"common.PrintProcessor": {}}], "input": 1}

==========================================
Curl call to the server with CHAP pipeline
==========================================
curl -X POST -H "Content-type: application/json" -d@/tmp/chap.json http://localhost:5000/pipeline
{"pipeline": [{"common.PrintProcessor":{}}], "status":"ok"}

===========
Server side
===========
flask --app server run

* Serving Flask app 'server'
* Debug mode: off

WARNING: This is a development server. Do not use it in a production
deployment. Use a production WSGI server instead.

* Running on http://127.0.0.1:5000

Press CTRL+C to quit

CHAP output::

    CHAP.server         : Call pipeline args=()
        kwds={'pipeline': [{'common.PrintProcessor': {}}]}
    CHAP.server         : pipeline [{'common.PrintProcessor': {}}]
    CHAP.server         : Loaded
        <CHAP.common.processor.PrintProcessor object at 0x10e0f1ed0>
    CHAP.server         : Loaded
        <CHAP.pipeline.Pipeline object at 0x10e0f1f10> with 1 items
    CHAP.server         : Calling "execute" on <CHAP.pipeline.Pipeline
        object at 0x10e0f1f10>
    Pipeline            : Executing "execute"
    Pipeline            : Calling "process" on
        <CHAP.common.processor.PrintProcessor object at 0x10e0f1ed0>
    PrintProcessor      : Executing "process" with
        type(data)=<class 'NoneType'>
    PrintProcessor data : None
    PrintProcessor      : Finished "process" in 0.000 seconds
    Pipeline            : Executed "execute" in 0.000 seconds
"""
# System modules
import time
from queue import Queue

# Third-party modules

# Flask modules
from flask import Flask, request

# Local modules
from CHAP.TaskManager import TaskManager, start_new_thread
from CHAP.runner import run, set_logger


# Task manager to execute our tasks
taskManager = TaskManager()

# Flask Server
app = Flask(__name__)

# daemon task queue
task_queue = Queue()

@app.route("/")
def index_route():
    """Server main end-point."""

    return "CHAP daemon"

@app.route("/run")
def run_route():
    """Server main end-point."""

    ttask = request.args.get('task')
    task_queue.put(ttask)
    return f'Execute {ttask}'

@app.route("/pipeline", methods=["POST"])
def pipeline_route():
    """Server /pipeline end-point."""

    content = request.json
    if 'pipeline' in content:
        # spawn new pipeline task
        jobs = []
        jobs.append(taskManager.spawn(task, pipeline=content['pipeline']))
        taskManager.joinall(jobs)
        return {'status': 'ok', 'pipeline': content['pipeline']}
    return {'status': 'fail', 'reason': 'no pipeline in incoming request'}

def task(*args, **kwds):
    """Helper function to execute CHAP pipeline."""

    log_level = 'INFO'
    logger, log_handler = set_logger(log_level)
    logger.info(f'call pipeline args={args} kwds={kwds}')
    pipeline = kwds['pipeline']
    logger.info(f'pipeline\n{pipeline}')
    run(pipeline, logger, log_level, log_handler)

def daemon(name, queue, interval):
    """Daemon example based on Queue."""

    print(f'Daemon {name}')
    while True:
        if queue.qsize() == 0:
            print('Default action')
            time.sleep(interval)
        else:
            ttask = queue.get()
            if ttask == 'exit':
                return
            print(f'daemon run {ttask}')

# start daemon thread in addition to Flask server
start_new_thread('daemon', daemon, ('daemon', task_queue, 3))
